# -*- coding: utf-8 -*-
"""Defines custom controls which re-render on resize."""
from __future__ import annotations

from typing import Any, Dict, Optional, Type

from prompt_toolkit.cache import SimpleCache
from prompt_toolkit.layout.controls import GetLinePrefixCallable, UIContent, UIControl

from euporie.render import (
    DataRenderer,
    HTMLRenderer,
    ImageRenderer,
    RichRenderer,
    SVGRenderer,
)
from euporie.text import ANSI


class Control(UIControl):
    """Base class for rich cell output.

    Will re-render it's output when the display is resized. Output is generated by
    `Control.renderer`, and is cached per output size.
    """

    renderer: "Type[DataRenderer]"

    def __init__(self, data: "Any", render_args: "Optional[Dict]" = None) -> "None":
        """Initalize the control.

        Args:
            data: Raw cell output data
            render_args: Additional keyword arguments to pass to the renderer.
        """
        self.data = data
        if render_args is None:
            render_args = {}
        self.render_args = render_args

        self.renderer_instance = self.renderer.select()
        self.rendered_lines: "list" = []
        self._format_cache: SimpleCache = SimpleCache(maxsize=20)
        self._content_cache: SimpleCache = SimpleCache(maxsize=20)

    def preferred_height(
        self,
        width: "int",
        max_available_height: "int",
        wrap_lines: "bool",
        get_line_prefix: Optional[GetLinePrefixCallable],
    ) -> "int":
        """Returns the number of lines in the rendered content."""
        if not self.rendered_lines:
            self.rendered_lines = self._format_cache.get(
                (width, max_available_height),
                lambda: self.render(width, max_available_height),
            )
        return len(self.rendered_lines)

    def create_content(self, width: "int", height: "int") -> "UIContent":
        """Generates rendered output at a given size.

        Args:
            width: The desired output width
            height: The desired output height

        Returns:
            `UIContent` for the given output size.

        """
        self.rendered_lines = self._format_cache.get(
            (width,),
            lambda: self.render(width, height),
        )

        def get_content() -> Optional[UIContent]:
            if self.rendered_lines is not None:
                return UIContent(
                    get_line=lambda i: ANSI(
                        self.rendered_lines[i]
                    ).__pt_formatted_text__(),
                    line_count=len(self.rendered_lines),
                )
            else:
                return None

        return self._content_cache.get((width,), get_content)

    def render(self, width: "int", height: "int") -> "list[str]":
        """Calls the renderer."""
        result = self.renderer_instance.render(
            self.data, width=width, height=height, render_args=self.render_args
        )
        rendered_lines = result.rstrip().split("\n")
        return rendered_lines


class RichControl(Control):
    """Control for rich renderables."""

    renderer = RichRenderer


class HTMLControl(Control):
    """Control for rendered HTML."""

    renderer = HTMLRenderer


class BaseImageControl(Control):
    """Base class for image controls."""

    def create_content(self, width: "int", height: "int") -> "UIContent":
        """Additionally cache rendered content by cell obscurity status."""
        cell_obscured = self.render_args["cell"].obscured()
        self.rendered_lines: "list" = self._format_cache.get(
            (cell_obscured, width),
            lambda: self.render(width, height),
        )

        def get_content() -> UIContent:
            return UIContent(
                get_line=lambda i: ANSI(self.rendered_lines[i]).__pt_formatted_text__(),
                line_count=len(self.rendered_lines),
            )

        return self._content_cache.get((cell_obscured, width), get_content)


class ImageControl(BaseImageControl):
    """Control for rendered raster images."""

    renderer = ImageRenderer


class SVGControl(BaseImageControl):
    """Class for rendered SVG iamges."""

    renderer = SVGRenderer



# -*- coding: utf-8 -*-
"""Contains dpdated ANSI parsing and Formatted Text processing."""
from __future__ import annotations

import re
from typing import Any, Generator

from prompt_toolkit.formatted_text import ANSI as PTANSI
from prompt_toolkit.formatted_text import (
    AnyFormattedText,
    FormattedText,
    fragment_list_to_text,
    split_lines,
    to_formatted_text,
)
from prompt_toolkit.layout.margins import ScrollbarMargin
from prompt_toolkit.layout.processors import (
    Processor,
    Transformation,
    TransformationInput,
)
from prompt_toolkit.widgets import TextArea


class FormatTextProcessor(Processor):
    """Applies formatted text to a TextArea."""

    def __init__(self, formatted_text: "FormattedText"):
        """Initiate the processor.

        Args:
            formatted_text: The text in a buffer but with formatting applied.

        """
        self.formatted_text = formatted_text
        super().__init__()

    def apply_transformation(
        self, transformation_input: "TransformationInput"
    ) -> "Transformation":
        """Apply text formatting to a line in a buffer."""
        if not hasattr(self, "formatted_lines"):
            self.formatted_lines = list(split_lines(self.formatted_text))
        lineno = transformation_input.lineno
        max_lineno = len(self.formatted_lines) - 1
        if lineno > max_lineno:
            lineno = max_lineno
        line = self.formatted_lines[lineno]
        return Transformation(line)


class FormattedTextArea(TextArea):
    """Applies formatted text to a TextArea."""

    def __init__(
        self, formatted_text: "AnyFormattedText", *args: "Any", **kwargs: "Any"
    ):
        """Initialise a `FormattedTextArea` instance.

        Args:
            formatted_text: A list of `(style, text)` tuples to display.
            *args: Arguments to pass to `prompt_toolkit.widgets.TextArea`.
            **kwargs: Key-word arguments to pass to `prompt_toolkit.widgets.TextArea`.

        """
        self.formatted_text = to_formatted_text(formatted_text)
        input_processors = kwargs.pop("input_processors", [])
        input_processors.append(FormatTextProcessor(self.formatted_text))
        text = fragment_list_to_text(self.formatted_text)
        kwargs.pop("text", "")

        # The following is not type checked due to a currently open mypy bug
        # https://github.com/python/mypy/issues/6799
        super().__init__(
            *args,
            text=text,
            input_processors=input_processors,
            **kwargs,
        )  # type: ignore

        for margin in self.window.right_margins:
            if isinstance(margin, ScrollbarMargin):
                margin.up_arrow_symbol = "▲"
                margin.down_arrow_symbol = "▼"


class ANSI(PTANSI):
    """Converts ANSI text into formatted text, preserving all control sequences."""

    def __init__(self, value: "str") -> None:
        """Initiate the ANSI processor instance.

        This replaces carriage returns to emulate terminal output.

        Args:
            value: The ANSI string to process.

        """
        # Replace windows style newlines
        value = value.replace("\r\n", "\n")
        # Remove anything before a carriage return if there is something after it to
        # emulate a carriage return in the output
        value = re.sub("^.*\\r(?!\\n)", "", value, 0, re.MULTILINE)

        super().__init__(value)

    def _parse_corot(self) -> Generator[None, str, None]:
        """Coroutine that parses the ANSI escape sequences.

        This is modified version of the ANSI parser from prompt_toolkit retains
        all CSI escape sequences.

        Yields:
            Accepts characters from a string.

        """
        style = ""
        formatted_text = self._formatted_text

        while True:

            char = yield
            sequence = char

            # Everything between \001 and \002 should become a ZeroWidthEscape.
            if char == "\001":
                sequence = ""
                while char != "\002":
                    char = yield
                    if char == "\002":
                        formatted_text.append(("[ZeroWidthEscape]", sequence))
                        break
                    else:
                        sequence += char
                continue

            # Check for backspace
            elif char == "\x08":
                # TODO - remove last character from last non-ZeroWidthEscape fragment
                formatted_text.pop()
                continue

            elif char in ("\x1b", "\x9b"):

                # Got a CSI sequence, try to compile a control sequence
                char = yield

                # Check for sixels
                if char == "P":
                    # Got as DEC code
                    sequence += char
                    # We expect "p1;p2;p3;q" + sixel data + "\x1b\"
                    char = yield
                    while char != "\x1b":
                        sequence += char
                        char = yield
                    sequence += char
                    char = yield
                    if ord(char) == 0x5C:
                        sequence += char
                        formatted_text.append(("[ZeroWidthEscape]", sequence))
                        # char = yield
                        continue

                # Check for hyperlinks
                elif char == "]":
                    sequence += char
                    char = yield
                    if char == "8":
                        sequence += char
                        char = yield
                        if char == ";":
                            sequence += char
                            char = yield
                            while True:
                                sequence += char
                                if sequence[-2:] == "\x1b\\":
                                    break
                                char = yield
                            formatted_text.append(("[ZeroWidthEscape]", sequence))
                            continue

                elif (char == "[" and sequence == "\x1b") or sequence == "\x9b":
                    if sequence == "\x1b":
                        sequence += char
                        char = yield

                    # Next are any number (including none) of "parameter bytes"
                    params = []
                    current = ""
                    while 0x30 <= ord(char) <= 0x3F:
                        # Parse list of integer parameters
                        sequence += char
                        if char.isdigit():
                            current += char
                        else:
                            params.append(min(int(current or 0), 9999))
                            if char == ";":
                                current = ""
                        char = yield
                    if current:
                        params.append(min(int(current or 0), 9999))
                    # then any number of "intermediate bytes"
                    while 0x20 <= ord(char) <= 0x2F:
                        sequence += char
                        char = yield
                    # finally by a single "final byte"
                    if 0x40 <= ord(char) <= 0x7E:
                        sequence += char
                    # Check if that escape sequence was a style:
                    if char == "m":
                        self._select_graphic_rendition(params)
                        style = self._create_style_string()
                    # Otherwise print a zero-width control sequence
                    else:
                        formatted_text.append(("[ZeroWidthEscape]", sequence))
                    continue

            formatted_text.append((style, sequence))



# -*- coding: utf-8 -*-
"""Define box border constants."""


class Border:
    """Box drawing characters, includeing characters for splits."""

    HORIZONTAL = "─"
    VERTICAL = "│"
    TOP_LEFT = "╭"
    TOP_RIGHT = "╮"
    BOTTOM_LEFT = "╰"
    BOTTOM_RIGHT = "╯"
    SPLIT_BOTTOM = "┴"
    SPLIT_TOP = "┬"
    SPLIT_LEFT = "├"
    SPLIT_RIGHT = "┤"
    CROSS = "┼"



# -*- coding: utf-8 -*-
"""Contains the main Application class which runs euporie."""
from __future__ import annotations

import asyncio
import io
import logging
import sys
from typing import IO, TYPE_CHECKING, Any, Callable, Optional, TextIO, Union, cast

from prompt_toolkit import renderer
from prompt_toolkit.application import Application
from prompt_toolkit.data_structures import Point, Size
from prompt_toolkit.output.defaults import create_output
from prompt_toolkit.output.vt100 import Vt100_Output
from prompt_toolkit.widgets import Box, HorizontalLine

from euporie.config import config

if TYPE_CHECKING:
    from prompt_toolkit.layout.containers import AnyContainer
    from prompt_toolkit.output import Output

log = logging.getLogger(__name__)

_original_output_screen_diff = renderer._output_screen_diff


def _patched_output_screen_diff(
    *args: "Any", **kwargs: "Any"
) -> "tuple[Point, Optional[str]]":

    # Remove ZWE from screen
    # from collections import defaultdict
    # args[2].zero_width_escapes = defaultdict(lambda: defaultdict(lambda: ""))

    # Tell the renderer we have one additional column. This is to prevent the use of carriage
    # returns and cursor movements to write the final character on lines.
    size = kwargs.pop("size")
    kwargs["size"] = Size(99999, size.columns + 1)
    return _original_output_screen_diff(*args, **kwargs)


class DumpMixin:
    """Provides user interface."""

    # Properties from `App` referred to here
    pre_run: "list[Callable]"
    files: "list[AnyContainer]"
    close_file: "Callable"
    exit: "Callable"
    output: "Output"
    out_file: "IO[str]"

    drawn = False

    def _redraw(self, render_as_done: "bool" = False) -> "None":
        """Ensure the output is drawn once, and the cursor is left after the output."""
        if not self.drawn:
            cast("Application", super())._redraw(render_as_done=True)
            self.drawn = True

    def setup(self) -> "None":
        """Patch the renderer and output to print files longer than the screen."""
        # Patch the renderer to extend the output height
        renderer._output_screen_diff = _patched_output_screen_diff

        if config.page and sys.stdout.isatty():
            # Use a temporary file as display output if we are going to page the output
            from tempfile import TemporaryFile

            self.out_file = TemporaryFile("w+")

        else:
            if config.page:
                log.warning("Cannot page output because standard output is not a TTY")
            # If we are not paging output, determine when to print it
            if config.dump_file is None or str(config.dump_file) in (
                "-",
                "/dev/stdout",
            ):
                self.out_file = sys.stdout
            elif str(config.dump_file) == "/dev/stderr":
                self.out_file = sys.stderr
            else:
                try:
                    self.out_file = open(config.dump_file, "w+")
                except (
                    FileNotFoundError,
                    PermissionError,
                    io.UnsupportedOperation,
                ) as error:
                    log.error(error)
                    log.error(
                        f"Output file `{config.dump_file}` cannot be opened. "
                        "Standard output will be used."
                    )
                    self.out_file = sys.stdout

        # Ensure we do not recieve the "Output is not a terminal" message
        Vt100_Output._fds_not_a_terminal.add(self.out_file.fileno())
        # Do not use stderr instead of stdout if stdout is not a tty
        self.out_file = cast("TextIO", self.out_file)
        self.output = create_output(self.out_file, always_prefer_tty=False)

        # Use the width and height of stderr (this gives us the terminal size even if
        # output is being piped to a non-tty)
        # if hasattr(self.output, '_get_size'):
        setattr(self.output, "get_size", create_output(stdout=sys.stderr).get_size)

        # Disable character position requests when dumping output to stop extra output
        # This also speeds things up as we do not need to wait for the response
        # Ignore typing here as mypy does not understand __class__
        class DumpingOutput(self.output.__class__):  # type: ignore
            # Disable character position requests when dumping output
            responds_to_cpr = False

        # Patch the output to prevent CPR detection
        self.output.__class__ = DumpingOutput

        # Set pre-run commands
        self.pre_run.append(self.post_dump)

    def post_dump(self) -> "None":
        """Close all files and exit the app immediately when dumping files."""
        # Close all the file immediately
        for file in self.files:
            self.close_file(file)

        # Queue exiting the application
        asyncio.get_event_loop().call_soon(self.pre_exit)

    def pre_exit(self) -> "None":
        """Close the app after dumping file, optionally piping output to a pager."""
        # Display pager
        if config.page:
            from pydoc import pager

            self.out_file.seek(0)
            data = self.out_file.read()
            pager(data)

        if self.out_file is not sys.stdout:
            self.out_file.close()

        # Close the application
        self.exit()

    def layout_container(self) -> "AnyContainer":
        """Returns a container with all opened files."""
        from euporie.scroll import PrintingContainer

        # Create a horizontal line that takes up the full width of the display
        hr = HorizontalLine()
        hr.window.width = self.output.get_size().columns

        # Add files, separated by horizontal lines
        contents: "list[Union[Callable, AnyContainer]]" = []
        for file in self.files:
            # Wrap each file in a box so it does not expand beyond its maximum width
            contents.append(Box(file))
            contents.append(hr)
        # Remove the final horizontal line
        if self.files:
            contents.pop()
        return Box(PrintingContainer(contents))

    def cpr_not_supported_callback(self) -> "None":
        """Hide messages about cursor position requrests."""
        pass



# -*- coding: utf-8 -*-
"""Adds markdown table support to rich.markdown.

Currently every table cell is treated as a paragraph, which makes the table take the
full width of the display.

TODO - fix this.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

import rich.markdown
from rich.console import Console, ConsoleOptions, JustifyMethod, RenderResult
from rich.table import Table as RichTable

from euporie.config import config
from euporie.libs.commonmark_extensions.tables import ParserWithTables  # type: ignore
from euporie.libs.commonmark_extensions.tables import Table as CETable  # type: ignore

if TYPE_CHECKING:
    from commonmark.node import Node  # type: ignore


rich.markdown.Parser = ParserWithTables


class Table(rich.markdown.MarkdownElement):
    """A table."""

    style_name = "markdown.table"
    new_line = False

    @classmethod
    def create(cls, markdown: "rich.markdown.Markdown", table: "CETable") -> "Table":
        """Instantiates and returns a rich markdown table."""
        return cls(table)

    def __init__(self, contents: "CETable") -> None:
        """Sets the contents for this markdown table.

        Args:
            contents: A `commonmark_extensions.tables.Table` instance.

        """
        self.contents = contents.table
        self.column_properties = contents.column_properties

    def __rich_console__(
        self, console: "Console", options: "ConsoleOptions"
    ) -> RenderResult:
        """Render the table instance.

        Args:
            console: The `rich.console.Console` instance to use to render this table.
            options: The console options to use.

        Yields:
            A rich table renderable.

        """
        table = RichTable()

        def node_to_text(
            node: "Node",
            justify: "JustifyMethod" = None,
        ) -> "rich.markdown.Markdown":
            """Create a rich markdown object, then add the parsed node to it."""
            node_md = rich.markdown.Markdown(
                "",
                code_theme=str(config.syntax_theme),
                justify=justify,
            )
            node_md.parsed = node
            return node_md

        for i, col in enumerate(self.contents[0][0]):
            justify = self.column_properties.get(i, {}).get("align")
            table.add_column(node_to_text(col, justify=justify), justify=justify)

        for row in self.contents[1]:
            table.add_row(*[node_to_text(cell, justify=justify) for cell in row])

        yield table


rich.markdown.Markdown.elements["table"] = Table



# -*- coding: utf-8 -*-
# type: ignore
# flake8: noqa
# Mokey-patch the reMaybeSpecial regex to add our table symbol |.
# This regex is apparently just an optimization so this should not
# affect CommonMark parser instances that do not recognize tables.
import re
from collections import defaultdict

import commonmark
import commonmark.blocks
import commonmark.node

commonmark.blocks.reMaybeSpecial = re.compile(r"^[#`~*+_=<>0-9-|]")


# Define a new BlockStarts class that implements a table method
# to detect and parse table starts, modeled after the blockquote.
class BlockStarts(commonmark.blocks.BlockStarts):
    def __init__(self):
        self.METHODS = ["table"] + self.METHODS

    @staticmethod
    def table(parser, container):
        if (
            not parser.indented
            and commonmark.blocks.peek(parser.current_line, parser.next_nonspace) == "|"
        ):
            parser.advance_next_nonspace()
            parser.advance_offset(1, False)
            parser.close_unmatched_blocks()
            parser.add_child("table", parser.next_nonspace)
            parser.tip.string_content = "|"
            return 2
        return 0


# Define a new Table class that handles incoming table lines, modeled
# a bit after the Blockquote, which allows continuation lines so long as
# they start with the symbol. Also has accepts_lines to suck in everything
# within it as raw data. Accept : as a continuation symbol for
# Github-flavored Markdown table column alignment.
class Table(commonmark.blocks.Block):
    accepts_lines = True

    @staticmethod
    def continue_(parser=None, container=None):
        ln = parser.current_line
        if (
            not parser.indented
            and commonmark.blocks.peek(ln, parser.next_nonspace) == "|"
        ):
            parser.advance_next_nonspace()
            parser.advance_offset(1, False)
            container.string_content += "|"
        elif not parser.indented and commonmark.blocks.peek(
            ln, parser.next_nonspace
        ) not in ("", ">", "`", None):
            pass
        else:
            return 1
        return 0

    @staticmethod
    def finalize(parser=None, block=None):
        # Split the table content into rows and columns,
        # with each line a new row.
        #
        # Note that the | in the first column is not a
        # part of string_content because it was removed
        # when we slurped in the table.
        table = [[""]]
        escape = False
        newrowbars = False  # noqa F841
        ignore_pipe = True
        for c in block.string_content.rstrip():
            # \-escaping
            if escape:
                table[-1][-1] += c
                escape = False
            elif c == "\\":
                escape = True

            # New cell is begun by a bar. Right-strip the cell we're
            # ending.
            elif c == "|":
                if ignore_pipe:
                    ignore_pipe = False
                    continue
                table[-1][-1] = table[-1][-1].rstrip()
                table[-1].append("")

            # New row is begun by a newline.
            # Since there's always a pipe at the end of a line,
            # pop last cell if empty.
            elif c == "\n":
                if table[-1][-1].strip() == "":
                    table[-1].pop(-1)
                table.append([""])
                ignore_pipe = True
                continue

            # Ignore space at start of cell. An escaped space
            # can force a space.
            elif c in (" ", "\t") and table[-1][-1] == "":
                pass

            # Content.
            else:
                table[-1][-1] += c

            ignore_pipe = False

        # Remove the last cell if it's empty since it's caused
        # by the final pipe at the end of the last line.
        if table[-1][-1].strip() == "":
            table[-1].pop(-1)

        # Re-flow the table into a <thead> part and a <tbody> part,
        # and if the separator row uses ='s instead of -'s then
        # treat subsequent rows as multiline rows that must be
        # separated by ='s.
        column_properties = defaultdict(lambda: {})
        table_parts = [[]]  # [thead, tbody] or just [tbody]
        multiline = False
        newrow = False
        for row in table:
            if len(list(filter(lambda cell: not re.match(r"[-=:]+$", cell), row))) == 0:
                # This row has cells of just dahses.
                if len(table_parts) == 1:
                    if len(row) != len(table_parts[0][0]):
                        block.t = "paragraph"
                        return
                    # The first time, we shift to the tbody.
                    table_parts.append([])

                    # We also pick out column properties from the
                    # placement of a colon.
                    for i, cell in enumerate(row):
                        if cell.startswith(":") and cell.endswith(":"):
                            column_properties[i]["align"] = "center"
                        elif cell.startswith(":"):
                            column_properties[i]["align"] = "left"
                        elif cell.endswith(":"):
                            column_properties[i]["align"] = "right"

                    # If ='s were used, then the table is parsed in
                    # multiline mode.
                    if "=" in "".join(row):
                        multiline = True

                elif multiline:
                    # Subsequent times we just note that we're starting a new row
                    # in multiline mode.
                    newrow = True

            elif not multiline or newrow or len(table_parts[-1]) == 0:
                # Append a new row.
                table_parts[-1].append(row)
                newrow = False

                # Fill in empty rows if fewer than the header.
                if len(table_parts) > 1 and len(table_parts[0][0]) > len(
                    table_parts[-1][-1]
                ):
                    table_parts[-1][-1].extend(
                        [
                            ""
                            for _ in range(
                                len(table_parts[0][0]) - len(table_parts[-1][-1])
                            )
                        ]
                    )

                # Remove excess cells if more than number of columns
                if len(table_parts) > 1 and len(table_parts[0][0]) < len(
                    table_parts[-1][-1]
                ):
                    table_parts[-1][-1] = table_parts[-1][-1][: len(table_parts[0][0])]

            else:
                # Multline mode. Merge this row with the previous one.
                for i in range(len(row)):
                    if i < len(table_parts[-1][-1]):
                        table_parts[-1][-1][i] += "\n" + row[i]
                    else:
                        table_parts[-1][-1].append(row[i])

        # Remove the last table part (probably tbody) if there is no content.
        if table_parts[-1] == []:
            table_parts.pop(-1)

        # Parse the Markdown in each cell using a new parser
        # instance for each cell.
        if not multiline:
            # Just parse the inlines in each cell using the parser's
            # inline_parser function. Wrap each cell string content
            # in a Node first.
            def inner_parser(cell):
                node = commonmark.node.Node("document", 0)
                para = commonmark.node.Node("paragraph", 0)
                para.string_content = cell
                node.append_child(para)
                parser.inline_parser.parse(para)
                return node

        else:
            # Parse each cell using the full Markdown parser,
            # by instantiating a new instance of the same
            # parser class first, using the same options.
            inner_parser = type(parser)(options=parser.options).parse

        for part in table_parts:
            for row in part:
                for i, cell in enumerate(row):
                    row[i] = inner_parser(cell)

        # Store the parsed table on the node.
        block.column_properties = column_properties
        block.table = table_parts


commonmark.blocks.Table = Table


# Create a new parser sub-class that adds the new block-start
# for tables.
class ParserWithTables(commonmark.Parser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.block_starts = BlockStarts()
        if hasattr(self, "blocks"):
            self.blocks["table"] = commonmark.blocks.Table


# Define a new renderer that extends the HtmlRenderer and
# adds table rendering.
class RendererWithTables(commonmark.HtmlRenderer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def make_table_node(self, node):
        return "<table>"

    def table(self, node, entering):
        if entering:
            self.lit(self.make_table_node(node) + "\n")
            for i, part in enumerate(node.table):
                if i == 0:
                    part_tag = "thead"
                else:
                    part_tag = "tbody"
                self.lit("<" + part_tag + ">\n")
                for row in part:
                    self.lit("<tr>\n")
                    for colidx, cell in enumerate(row):
                        if part_tag == "thead":
                            col_tag = "th"
                            if self.options.get("table_th_scope"):
                                col_attrs = '  scope="col"'
                            else:
                                col_attrs = ""
                        else:
                            col_tag = "td"
                            col_attrs = ""

                        if (
                            colidx in node.column_properties
                            and "align" in node.column_properties[colidx]
                        ):
                            col_attrs += (
                                ' align="'
                                + node.column_properties[colidx]["align"]
                                + '"'
                            )

                        self.lit("<" + col_tag + col_attrs + ">")

                        import copy

                        inner_renderer = copy.copy(self)
                        cell = inner_renderer.render(cell)

                        # If the cell is just one <p>, unwrap it.
                        m = re.match("<p>(.*)</p>$", cell)
                        if m:
                            cell = m.group(1)

                        self.lit(cell)
                        self.lit("</" + col_tag + ">\n")
                    self.lit("</tr>\n")
                self.lit("</" + part_tag + ">\n")
            self.lit("</table>\n")


# Define a new helper method that would be an in-place replacement
# for commonmark.commonmark.
def commonmark_to_html(markup):
    parser = ParserWithTables()
    ast = parser.parse(markup)
    return RendererWithTables().render(ast)


if __name__ == "__main__":
    # Run the parser on STDIN and write to STDOUT.
    import sys

    parser = ParserWithTables()
    ast = parser.parse(sys.stdin.read())
    print(RendererWithTables().render(ast))



# -*- coding: utf-8 -*-
"""Contains classes related to querying terminal feautres."""
from __future__ import annotations

import array
import re
import sys
from functools import lru_cache
from typing import IO, Callable, Optional, cast

from prompt_toolkit.output import Output

CONTROL_RE = re.compile(
    r"""
\x1b
(?:
 (?P<csi>  # Control Sequence Introducer
  \[
  (?P<csi_params>[0-9:;<=>?]*)
  (?P<csi_inter>[!\"#$%&'()*+,\-./]*)
  (?P<csi_final>[A-Za-z@[\]^_`a–z{|}~])
 )
 |
 (?P<osc>  # Operating System Command
  \]
  (?P<osc_string>.*(?=\x1b\\|\x07))
  (?P<osc_term>\x1b\\|\x07) # Terminated by ST or BEL
 )
 |
 (?P<apc>  # Application Program Command
  _
  (?P<apc_string>.*(?=\x1b\\))
  (?P<apc_term>\x1b\\) # Terminated by ST or BEL
 )
 |
 (?P<fe> # Other Fe escape sequences
  [NOP[\\\]X^_]
 )
)
    """,
    re.VERBOSE,
)


class QueryCodes:
    """Container for terminal query strings."""

    device = "\x1b[>0c"
    pixel_dimensions = "\x1b[14t"
    background_color = "\x1b]11;?\x1b\\"
    sixel = "\033[?1;1;0S"
    # For some reason, konsole prints APC codes to the screen, so we clear the current
    # line and move the curor to the start after sending a kitty graphics query code,
    # which prevents breaking the display
    kitty = "\x1b_Gi=1,s=1,v=1,a=q,t=d,f=24;AAAA\x1b\\" + "\x1b[1K" + "\x1b[1G"


class QueryResponsePatterns:
    """Container for terminal query response patterns."""

    device = re.compile(r"\x1b\[\>(\d;?)+c")
    background_color = re.compile(
        "11;rgb:(?P<r>[0-9A-Fa-f]{2,4})/(?P<g>[0-9A-Fa-f]{2,4})/(?P<b>[0-9A-Fa-f]{2,4})"
    )
    pixel_dimensions = re.compile(r"4;(?P<y>\d+);(?P<x>\d+)")
    sixel = re.compile(r"\d+")
    kitty = "\x1b[_Gi=1;OK\x1b[\\"


@lru_cache
def _have_termios_tty_fcntl() -> "bool":
    try:
        import fcntl  # noqa F401
        import termios  # noqa F401
        import tty  # noqa F401
    except ModuleNotFoundError:
        return False
    else:
        return True


def _query_term(
    query_code: "str", stdin: "IO[str]" = None, stdout: "IO[str]" = None
) -> "Optional[dict[str,str]]":
    """Query the terminal and parse the response.

    A Secondary Device Attribute request is used as the deliminator to locate the
    query response string.

    Args:
        query_code: The ANSI escape sequence to send to the terminal.
        stdin: IO to use as stdin
        stdout: IO to use as stdout

    Returns:
        A parsed escape code response.

    """
    if stdin is None:
        stdin = sys.stdin
    if stdout is None:
        stdout = sys.stdout

    # Check for tty in stdout
    if hasattr(stdout, "isatty"):
        if not stdout.isatty():
            return None

    if _have_termios_tty_fcntl():
        import termios
        import tty

        stdout_write: "Callable"
        if isinstance(stdout, Output):
            stdout_write = cast("Output", stdout).write_raw
        else:
            stdout_write = stdout.write

        tattr = termios.tcgetattr(stdin)

        output = ""

        try:
            tty.setcbreak(stdin, termios.TCSANOW)
            stdout_write(QueryCodes.device)
            stdout.flush()

            output_started = False
            query_sent = False

            while True:
                output += stdin.read(1)
                # Send query after first character of first deliminator recieved
                if not query_sent:
                    stdout_write(query_code)
                    stdout.flush()
                    query_sent = True
                if re.search(QueryResponsePatterns.device, output):
                    if not output_started:
                        output_started = True
                        # Discard the first deliminator
                        output = ""
                        # Send the second deliminator
                        stdout_write(QueryCodes.device)
                        stdout.flush()
                        continue
                    break
        finally:
            termios.tcsetattr(stdin, termios.TCSANOW, tattr)
        # Remove deliminator
        output = re.sub(QueryResponsePatterns.device, "", output, 1)
        # Parse result
        if match := CONTROL_RE.search(output):
            return match.groupdict()
    return None


def _tiocgwnsz() -> "tuple[int, int, int, int]":
    """Get the size and pixel dimensions of the terminal with `termios`."""
    output = array.array("H", [0, 0, 0, 0])
    if _have_termios_tty_fcntl():
        import fcntl
        import termios

        try:
            fcntl.ioctl(1, termios.TIOCGWINSZ, output)
        except Exception:  # noqa S110
            pass
    rows, cols, xpixels, ypixels = output
    return rows, cols, xpixels, ypixels


class TermMixin:
    """Mixin for `euporie.app.App` which provides terminal feature properties."""

    output: "Output"

    @property
    def bg_color(self) -> "Optional[str]":
        """Get the background colour of the terminal as a hex colour code."""
        if result := _query_term(
            QueryCodes.background_color, stdout=self.output.stdout
        ):
            if oscs := result.get("osc_string"):
                if match := QueryResponsePatterns.background_color.match(oscs):
                    if colors := match.groupdict():
                        if len(colors) >= 3:
                            r, g, b = (
                                colors.get("r", "00"),
                                colors.get("g", "00"),
                                colors.get("b", "00"),
                            )
                            return f"#{r[:2]}{g[:2]}{b[:2]}"
        return None

    @property
    def term_size_px(self) -> "tuple[int, int]":
        """Get the pixel dimensions of the terminal."""
        *_, px, py = _tiocgwnsz()
        # If unsuccessful, try requesting info with escape code method
        if px == 0:
            if result := _query_term(
                QueryCodes.pixel_dimensions, stdout=self.output.stdout
            ):
                params = result.get("csi_params", "")
                if match := QueryResponsePatterns.pixel_dimensions.match(params):
                    if values := match.groupdict():
                        if (x := values.get("x") is not None) and (
                            y := values.get("y") is not None
                        ):
                            px, py = int(x), int(y)
        return px, py

    @property
    def char_size_px(self) -> "tuple[int, int]":
        """Get the pixel size of a single terminal character block."""
        px, py = self.term_size_px
        rows, cols = self.output.get_size()
        # If we can't get the pixel size, just guess wildly
        return px // cols or 10, py // rows or 22

    @property  # type: ignore
    @lru_cache
    def has_sixel_graphics(self) -> "bool":
        """Determine if the terminal supports sixel graphics."""
        if result := _query_term(QueryCodes.sixel, stdout=self.output.stdout):
            param_str = result.get("csi_params", "")
            matches = QueryResponsePatterns.sixel.findall(param_str)
            params = [int(x) for x in matches]
            if len(params) >= 3:
                Pi, Ps, Pv = params[:3]
                if Ps == 0:
                    return True
        return False

    @property  # type: ignore
    @lru_cache
    def has_kitty_graphics(self) -> "bool":
        """Determine if the terminal supports the kitty graphics protocal."""
        if result := _query_term(QueryCodes.kitty, stdout=self.output.stdout):
            apc_string = result.get("apc_string", "")
            if apc_string.startswith("G"):
                if len(response := apc_string.lstrip("G").split(";")) >= 2:
                    if response[1] == "OK":
                        return True
        return False



# -*- coding: utf-8 -*-
"""Contains the main class for a notebook file."""
from __future__ import annotations

import asyncio
import copy
import threading
from functools import partial
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    AsyncGenerator,
    Callable,
    Iterable,
    Optional,
    Union,
    cast,
)

import nbformat  # type: ignore
from jupyter_client import KernelClient, KernelManager  # type: ignore
from jupyter_client.kernelspec import KernelSpecManager  # type: ignore
from prompt_toolkit.application.current import get_app
from prompt_toolkit.completion.base import CompleteEvent, Completer, Completion
from prompt_toolkit.document import Document
from prompt_toolkit.layout.containers import AnyContainer, HSplit, Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import LayoutDimension as D
from prompt_toolkit.layout.margins import NumberedMargin
from prompt_toolkit.widgets import Label, RadioList

from euporie.cell import Cell
from euporie.config import config
from euporie.keys import KeyBindingsInfo
from euporie.scroll import PrintingContainer, ScrollingContainer

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindings
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent

    from euporie.app import App


class File:
    """Base class for file containers."""

    def close(self, cb: "Optional[Callable]") -> "None":
        """Function to close a file.

        Args:
            cb: A function to call when the file is closed.

        """
        pass

    def focus(self) -> "None":
        """Focuses the file."""
        pass

    def __pt_container__(self) -> "AnyContainer":
        """Return the file container object."""
        pass


class KernelCompleter(Completer):
    """A prompt_toolkit completer which provides completions from a Jupyter kernel."""

    def __init__(self, nb: "Notebook"):
        """Instantiate the completer for a given notebook.

        Args:
            nb: A `Notebook` instance

        """
        self.nb = nb

    def get_completions(
        self, document: "Document", complete_event: "CompleteEvent"
    ) -> "Iterable[Completion]":
        """Does nothing as completions are retrieved asynchronously."""
        while False:
            yield

    async def get_completions_async(
        self, document: Document, complete_event: CompleteEvent
    ) -> "AsyncGenerator[Completion, None]":
        """An asynchronous generator of `Completions`, as returned by the kernel."""
        if self.nb.kc:
            msg_id = self.nb.kc.complete(
                code=document.text,
                cursor_pos=document.cursor_position,
            )
            # msg = await self.nb.kc._async_get_shell_msg()
            msg = self.nb.kc.get_shell_msg()
            if msg["parent_header"].get("msg_id") == msg_id:
                # run_in_terminal(lambda: print(msg))
                content = msg.get("content", {})
                rel_start_position = (
                    content.get("cursor_start", 0) - document.cursor_position
                )
                for match in content.get("matches", []):
                    yield Completion(match, start_position=rel_start_position)


class Notebook(File):
    """The main notebook container class."""

    page: "ScrollingContainer"

    def __init__(
        self,
        path: "Path",
        interactive: "bool" = True,
        execute: "bool" = False,
        scroll: "bool" = False,
    ):
        """Instantiate a Notebook container, using a notebook at a given path."""
        self.path = Path(path).expanduser()
        self.interactive = interactive
        self.execute = execute
        self.scroll = scroll

        # Open json file
        if self.path.exists():
            self.json = nbformat.read(self.path, as_version=4)
        else:
            self.json = nbformat.v4.new_notebook()
        # Ensure there is always at least one cell
        if not self.json.setdefault("cells", []):
            self.json["cells"] = [nbformat.v4.new_code_cell()]

        self.clipboard: "list[Cell]" = []
        self.dirty = False
        self.kernel_status = "starting"
        self.line_numbers = config.line_numbers
        self.completer = KernelCompleter(self)
        self.km: "Optional[KernelManager]" = None
        self.kc: "Optional[KernelClient]" = None
        self.kernel_loop: "Optional[asyncio.AbstractEventLoop]" = None

        self.container: "AnyContainer"

        if not self.scroll:
            self.container = PrintingContainer(
                self.cell_renderers,
                width=config.max_notebook_width,
            )

        else:
            self.page = ScrollingContainer(
                children=self.cell_renderers,
                max_content_width=D(preferred=int(config.max_notebook_width)),
            )
            # Wrap the scolling container in an hsplit and apply the keybindings
            # TODO - refactor key-bindings
            self.container = HSplit([self.page])
            self.container.key_bindings = self.load_key_bindings()

        def setup_loop() -> None:
            """Set up a thread with an event loop to listen for kernel responses."""
            # Create and set kernel loop
            self.kernel_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.kernel_loop)
            # Start kernel
            asyncio.run_coroutine_threadsafe(
                self.start_kernel(),
                self.kernel_loop,
            )
            # Now we wait
            self.kernel_loop.run_forever()

        # Run the kernel event loop in a new thread
        if self.interactive or self.execute:
            self.kernel_thread = threading.Thread(target=setup_loop)
            self.kernel_thread.start()

    async def start_kernel(self) -> None:
        """Starts a Juypter kernel, creating a `KernelManager`."""
        if not self.kernel_name:
            self.change_kernel()
            return
        # Create a kernel manager for this notebook
        self.km = KernelManager(kernel_name=self.kernel_name)
        try:
            # await self.km._async_start_kernel()
            self.km.start_kernel()
        except Exception as e:
            app = cast("App", get_app())
            app.dialog("Error Starting Kernel", Label(e.__repr__()), {"OK": None})
            self.kc = None
            self.kernel_status = "error"
        else:
            self.kc = self.km.client()
            self.kc.start_channels()
            self.kernel_status = "idle"
        get_app().invalidate()

    def restart_kernel(self) -> "None":
        """Restarts the current `Notebook`'s kernel.

        This is performed asynchronously in the notebook's kernel thread.
        """

        async def _restart() -> "None":
            if self.km:
                await self.km._async_shutdown_kernel()
            await self.start_kernel()

        assert self.kernel_loop is not None
        asyncio.run_coroutine_threadsafe(
            _restart(),
            self.kernel_loop,
        )

    @property
    def kernel_name(self) -> "str":
        """Return the name of the kernel defined in the notebook JSON."""
        return self.json.get("metadata", {}).get("kernelspec", {}).get("name")

    def change_kernel(self) -> None:
        """Displays a dialog for the user to select a new kernel."""

        def _change_kernel_cb() -> None:
            name = options.current_value
            spec = kernel_specs.get(name, {}).get("spec", {})
            self.json.setdefault("metadata", {})["kernelspec"] = {
                "display_name": spec["display_name"],
                "language": spec["language"],
                "name": name,
            }
            self.restart_kernel()

        if self.km:
            kernel_specs = self.km.kernel_spec_manager.get_all_specs()
        else:
            kernel_specs = KernelSpecManager().get_all_specs()
        options = RadioList(
            [
                (
                    kernel_name,
                    kernel_spec.get("spec", {}).get("display_name", kernel_name),
                )
                for kernel_name, kernel_spec in kernel_specs.items()
            ]
        )
        app = cast("App", get_app())
        app.dialog(
            title="Select Kernel",
            body=HSplit(
                [
                    Label("Please select a kernel:"),
                    options,
                ]
            ),
            buttons={
                "Select": _change_kernel_cb,
                "Cancel": None,
            },
        )

    def focus(self) -> "None":
        """Focus the notebooks."""
        if hasattr(self, "page"):
            self.page.focus()

    @property
    def cell_renderers(self) -> "list[Union[Callable, AnyContainer]]":
        """Return a list of `Cell` generator functions for the notebooks' cells."""
        return [
            partial(Cell, i, cell_json, self)
            for i, cell_json in enumerate(self.json["cells"])
        ]

    def get_cell_by_id(self, cell_id: "str") -> "Optional[Cell]":
        """Returns a reference to the `Cell` container with a given cell id."""
        for cell in self.page.child_cache.values():
            assert isinstance(cell, Cell)
            if cell.id == cell_id:
                break
        else:
            return None
        return cell

    def load_key_bindings(self) -> "KeyBindings":
        """Load the key bindings associate with a `Notebook` container."""
        kb = KeyBindingsInfo()

        @kb.add("c-s", group="Application", desc="Save current file")
        def save(event: "KeyPressEvent") -> "None":
            self.save()

        @kb.add("a", group="Notebook", desc="Add new cell above")
        def add_above(event: "KeyPressEvent") -> "None":
            self.add(0)

        @kb.add("b", group="Notebook", desc="Add new cell below")
        def add_below(event: "KeyPressEvent") -> "None":
            self.add(1)

        @kb.add("d", "d", group="Notebook", desc="Delete current cell")
        def delete(event: "KeyPressEvent") -> "None":
            self.delete()

        @kb.add("x", group="Notebook", desc="Cut current cell")
        def cut(event: "KeyPressEvent") -> "None":
            self.cut()

        @kb.add("c", group="Notebook", desc="Copy current cell")
        def copy(event: "KeyPressEvent") -> "None":
            self.copy()

        @kb.add("v", group="Notebook", desc="Paste copied cell")
        def paste(event: "KeyPressEvent") -> "None":
            self.paste()

        @kb.add("m", group="Notebook", desc="Change cell to markdown")
        def to_markdown(event: "KeyPressEvent") -> "None":
            self.cell.set_cell_type("markdown")
            self.cell.clear_output()

        @kb.add("y", group="Notebook", desc="Change cell to code")
        def to_code(event: "KeyPressEvent") -> "None":
            self.cell.set_cell_type("code")

        @kb.add("r", group="Notebook", desc="Change cell to raw")
        def to_raw(event: "KeyPressEvent") -> "None":
            self.cell.set_cell_type("raw")
            self.cell.clear_output()

        @kb.add("l", group="Notebook", desc="Toggle line numbers")
        def line_nos(event: "KeyPressEvent") -> "None":
            self.line_numbers = not self.line_numbers
            for cell in self.page.child_cache.values():
                assert isinstance(cell, Cell)
                cell.input_box.window.left_margins = (
                    [NumberedMargin()] if self.line_numbers else []
                )

        return kb

    @property
    def cell(self) -> "Cell":
        """Returns the currently selected `Cell` in this `Notebook`."""
        cell = self.page.get_child()
        assert isinstance(cell, Cell)
        return cell

    def is_cell_obscured(self, index: "int") -> "bool":
        """Determine if a cell is partially visible.

        Args:
            index: The index of the child of interest.

        Returns:
            True if the child is rendered and partially off-screen, otherwise False.=

        """
        if self.scroll:
            return self.page.is_child_obscured(index)
        else:
            return False

    def add(self, offset: "int") -> "None":
        """Creates a new cell at a given offset from the currently selected cell.

        Args:
            offset: How many cells above (negative) or below (positive) the current
                cell the new cell should be placed.

        """
        index = self.page.selected_index + offset
        self.json["cells"].insert(
            index,
            nbformat.v4.new_code_cell(),
        )
        self.dirty = True
        self.refresh(index=index)

    def cut(self, index: "Optional[int]" = None) -> "None":
        """Remove a cell from the notebook and add it to the `Notebook`'s clipboard."""
        if index is None:
            index = self.page.selected_index
        self.copy(index)
        self.delete(index)

    def copy(self, index: "Optional[int]" = None) -> "None":
        """Add a copy of this cell to the `Notebook`'s clipboard."""
        if index is None:
            index = self.page.selected_index
        self.clipboard = copy.deepcopy(self.json["cells"][index : index + 1])

    def paste(self, index: "Optional[int]" = None) -> "None":
        """Append the contents of the `Notebook`'s clipboard below the current cell."""
        if index is None:
            index = self.page.selected_index
        self.json["cells"][index + 1 : index + 1] = copy.deepcopy(self.clipboard)
        self.dirty = True
        self.refresh(index + 1)

    def delete(self, index: "Optional[int]" = None) -> "None":
        """Delete a cell from the notebook."""
        if len(self.json["cells"]) > 1:
            if index is None:
                index = self.page.selected_index
            del self.json["cells"][index]
            self.dirty = True
            self.refresh()

    def refresh(self, index: "Optional[int]" = None) -> "None":
        """Refresh the rendered contents of this notebook."""
        if index is None:
            index = self.page.selected_index
        self.page.children = self.cell_renderers
        self.page.reset()
        self.page._set_selected_index(index, force=True)

    def save(self) -> "None":
        """Write the notebook's JSON to the current notebook's file."""
        self.json = nbformat.from_dict(self.json)
        nbformat.write(nb=self.json, fp=self.path)
        self.dirty = False

    def close(self, cb: "Optional[Callable]") -> "None":
        """Check if the user want to save an unsaved notebook, then close the file.

        Args:
            cb: A callback to run if after closing the notebook.

        """
        if self.dirty:
            self.unsaved(cb)
        else:
            self.really_close(cb)

    def really_close(self, cb: "Optional[Callable]") -> "None":
        """Shutdown the kernel and close the notebook.

        Args:
            cb: A callback to run if after closing the notebook.

        """
        # Tell the kernel to shutdown
        if self.kc is not None:
            self.kc.shutdown()

        if self.kernel_loop is not None:
            # Tell the event loop in the kernel monitoring thread to shutdown
            self.kernel_loop.call_soon_threadsafe(self.kernel_loop.stop)
            # Close the kernel monitoring thread
            self.kernel_thread.join()

        # Tell the app we've closed
        if cb is not None:
            cb()

    def unsaved(self, cb: "Optional[Callable]") -> "None":
        """Display a dialog prompting the user to save unsaved changes.

        Args:
            cb: A callback to run if after closing the notebook.

        """
        app = cast("App", get_app())

        def yes_cb() -> None:
            self.save()
            self.really_close(cb)

        def no_cb() -> None:
            self.really_close(cb)

        app.dialog(
            title="Unsaved Changes",
            body=Window(
                FormattedTextControl(
                    [
                        ("bold", self.path.name),
                        ("", " has unsaved changes\n"),
                        ("", "Do you want to save your changes?"),
                    ]
                ),
                dont_extend_height=True,
            ),
            buttons={
                "Yes": yes_cb,
                "No": no_cb,
                "Cancel": None,
            },
        )

    def __pt_container__(self) -> "AnyContainer":
        """Return the main `Notebook` container object."""
        return self.container



# -*- coding: utf-8 -*-
"""Container class for cell output."""
from __future__ import annotations

from pathlib import PurePath
from typing import TYPE_CHECKING, Any

from prompt_toolkit.layout import Window
from prompt_toolkit.layout.controls import FormattedTextControl, UIControl
from prompt_toolkit.widgets import Label
from rich.markdown import Markdown

from euporie import mdtable  # noqa E401
from euporie.config import config
from euporie.control import HTMLControl, ImageControl, RichControl, SVGControl
from euporie.text import ANSI

if TYPE_CHECKING:
    from prompt_toolkit.layout.containers import AnyContainer

    from euporie.cell import Cell

BLING_SCORES = {
    "image/*": 0,
    "text/html": 1,
    "text/x-markdown": 2,
    "text/x-python-traceback": 3,
    "text/stderr": 4,
    "text/*": 5,
    "*": 6,
}


def calculate_bling(item: tuple[str, str]) -> int:
    """Scores the richness of mime output types."""
    mime, _ = item
    for bling_path, score in BLING_SCORES.items():
        if PurePath(mime).match(bling_path):
            return score
    else:
        return 999


class Output:
    """A prompt-toolkit compatible container for rendered cell outputs."""

    def __init__(self, index: "int", json: "dict[str, Any]", parent: "Cell"):
        """Instantiate an Output container object.

        Args:
            index: The position of this output in the list of the parent cell's outputs
            json: A reference to the notebook json branch corresponding to this output
            parent: A reference to the parent cell to which this output belongs
        """
        self.index = index
        self.json = json
        self.parent = parent
        self.style = ""
        self.content: AnyContainer

        control: UIControl

        # Sort data first so there is more bling first
        for mime, datum in sorted(self.data.items(), key=calculate_bling):

            mime_path = PurePath(mime)

            if mime_path.match("image/svg+xml"):
                control = SVGControl(
                    datum,
                    render_args=dict(
                        cell=self.parent,
                        cell_index=self.parent.index,
                        output_index=index,
                    ),
                )
                if control.renderer:
                    self.content = Window(control)
                    break

            if mime_path.match("image/*"):
                control = ImageControl(
                    datum,
                    render_args=dict(
                        cell=self.parent,
                        cell_index=self.parent.index,
                        output_index=index,
                    ),
                )
                if control.renderer:
                    self.content = Window(control)
                    break

            if mime_path.match("text/html"):
                control = HTMLControl(datum)
                if control.renderer:
                    self.content = Window(control)
                    break
                else:
                    continue  # Use plain text rendering instead

            if mime_path.match("text/x-markdown"):
                self.content = Window(
                    RichControl(
                        Markdown(
                            datum,
                            code_theme=str(config.syntax_theme),
                            inline_code_theme=str(config.syntax_theme),
                        )
                    )
                )
                break

            if mime_path.match("text/x-python-traceback"):
                self.content = Window(FormattedTextControl(ANSI(datum.rstrip())))
                break

            if mime_path.match("stream/std*"):
                # Parse ansi for text streams
                self.content = Window(
                    FormattedTextControl(ANSI(datum.rstrip())),
                    wrap_lines=True,
                    style="fg:red" if mime_path.name == "stderr" else "",
                )
                break

            if mime_path.match("text/*"):
                self.content = Window(
                    FormattedTextControl(datum.rstrip()), wrap_lines=True
                )
                break

        else:
            datum = sorted(self.data.items(), key=calculate_bling)[-1][1]
            self.content = Label(str(datum).rstrip())

    @property
    def data(self) -> "dict[str, str]":
        """Return dictionary of mime types and data for this output.

        This generates similarly structured data objects for markdown cells and text
        output streams.

        Returns:
            JSON dictionary mapping mimes type to representaion data.

        """
        output_type = self.json.get("output_type", "unknown")
        if output_type == "stream":
            return {f'stream/{self.json.get("name")}': self.json.get("text", "")}
        elif output_type == "error":
            return {
                "text/x-python-traceback": "\n".join(self.json.get("traceback", ""))
            }
        else:
            return self.json.get("data", {})

    def __pt_container__(self) -> "AnyContainer":
        """Return the content of this output."""
        return self.content



# -*- coding: utf-8 -*-
"""Instantiates the euporie package."""

__app_name__ = "euporie"
__version__ = "0.1.15"
__logo__ = "⚈"
__strapline__ = "A TUI editor for Jupyter notebooks"
__author__ = "Josiah Outram Halstead"
__email__ = "josiah@halstead.email"
__copyright__ = f"© 2021, {__author__}"
__license__ = "MIT"

import euporie.log  # noqa E401



# -*- coding: utf-8 -*-
"""Defines a cell object with input are and rich outputs, and related objects."""
from __future__ import annotations

import asyncio
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Literal

import nbformat  # type: ignore
from prompt_toolkit.application.current import get_app
from prompt_toolkit.buffer import Completion, indent, unindent
from prompt_toolkit.filters import Condition, has_completions, has_selection
from prompt_toolkit.formatted_text.base import StyleAndTextTuples
from prompt_toolkit.layout.containers import (
    ConditionalContainer,
    Container,
    Float,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
    to_container,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.lexers import DynamicLexer, PygmentsLexer
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.search import start_search
from prompt_toolkit.widgets import SearchToolbar, TextArea
from pygments.lexers import get_lexer_by_name  # type: ignore

from euporie.box import Border
from euporie.config import config
from euporie.keys import KeyBindingsInfo
from euporie.output import Output

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindings
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.layout import FocusableElement

    from euporie.notebook import Notebook


@Condition
def cursor_in_leading_ws() -> "bool":
    """Determine if the cursor of the current buffer is in leading whitespace."""
    before = get_app().current_buffer.document.current_line_before_cursor
    return (not before) or before.isspace()


class ClickArea:
    """Any empty widget which focuses `target` when clicked.

    Designed to be used as an overlay for clickable widgets in a FloatContainer.
    """

    def __init__(self, target: "FocusableElement"):
        """Initiate a click area overlay element, which focuses another element when clicked.

        Args:
            target: The element to focus on click.

        """
        self.target = target
        self.window = Window(
            FormattedTextControl(
                self._get_text_fragments,
                focusable=False,
            ),
            dont_extend_width=False,
            dont_extend_height=False,
        )

    def _get_text_fragments(self) -> "StyleAndTextTuples":
        def handler(mouse_event: MouseEvent) -> None:
            if mouse_event.event_type == MouseEventType.MOUSE_UP:
                get_app().layout.focus(self.target)

        return [("class:cell-clickarea", "", handler)]

    def __pt_container__(self) -> "Container":
        """Return the `ClickArea`'s window with a blank `FormattedTextControl`."""
        return self.window


class Cell:
    """A notebook cell element.

    Contains a transparent clickable overlay, which is not displayed when the cell is focused.
    """

    def __init__(self, index: "int", json: "dict", notebook: "Notebook"):
        """Initiate the cell element."""
        self.index = index
        self.json = json
        self.nb = notebook
        self.rendered = True
        self.editing = False

        self.state = "idle"

        ft = FormattedTextControl(
            Border.TOP_LEFT,
            focusable=True,
            show_cursor=False,
        )
        self.control = Window(ft, width=1, height=0, style=self.border_style)

        self.show_input = Condition(
            lambda: bool(
                (self.json.get("cell_type") != "markdown")
                | ((self.json.get("cell_type") == "markdown") & ~self.rendered)
            )
        )
        self.show_output = Condition(
            lambda: (
                (self.json.get("cell_type") != "markdown") & bool(self.outputs)
                | ((self.json.get("cell_type") == "markdown") & self.rendered)
            )
        )
        self.scroll_input = Condition(
            lambda: bool((self.json.get("cell_type") == "markdown") & ~self.rendered)
        )
        self.wrap_input = Condition(lambda: self.json.get("cell_type") == "markdown")
        self.is_editing = Condition(lambda: self.editing)
        self.show_prompt = Condition(lambda: self.cell_type == "code")
        self.is_focused = Condition(lambda: self.focused)
        self.obscured = Condition(lambda: self.nb.is_cell_obscured(self.index))
        self.show_input_line_numbers = Condition(
            lambda: bool(self.nb.line_numbers and self.json.get("cell_type") == "code")
        )

        self.input_box: TextArea
        self.container: FloatContainer
        self.output_box: HSplit

        self.load()

    def load_key_bindings(self) -> "KeyBindings":
        """Loads the key bindings related to cells."""
        kb = KeyBindingsInfo()

        @kb.add(
            "e", filter=~self.is_editing, group="Notebook", desc="Edit cell in $EDITOR"
        )
        async def edit_in_editor(event: "KeyPressEvent") -> "None":
            self.editing = True
            await self.input_box.buffer.open_in_editor()
            exit_edit_mode(event)
            if config.execute_after_external_edit:
                run_or_render(event)

        @kb.add(
            "enter",
            filter=~self.is_editing,
            group="Notebook",
            desc="Enter cell edit mode",
        )
        def enter_edit_mode(event: "KeyPressEvent") -> "None":
            self.editing = True
            self.container.modal = True
            get_app().layout.focus(self.input_box)
            self.rendered = False

        @kb.add("escape", group="Notebook", desc="Exit cell edit mode")
        @kb.add(
            "escape", "escape", group="Notebook", desc="Exit cell edit mode quickly"
        )
        def exit_edit_mode(event: "KeyPressEvent") -> "None":
            self.editing = False
            self.input = self.input_box.text
            self.nb.dirty = True
            self.container.modal = False
            # give focus back to selected cell (this might have changed!)
            get_app().layout.focus(self.nb.cell.control)

        @kb.add(
            "escape",
            "[",
            "1",
            "3",
            ";",
            "5",
            "u",
            key_str=("c-enter",),
            group="Notebook",
            desc="Run cell",
        )
        @kb.add("c-e", group="Notebook", desc="Run cell")
        @kb.add("c-f20")
        def run_or_render(event: "KeyPressEvent") -> "None":
            exit_edit_mode(event)
            if self.cell_type == "markdown":
                self.output_box.children = self.rendered_outputs
                self.rendered = True
            elif self.cell_type == "code":
                self.state = "queued"
                self.run()

        @kb.add(
            "escape",
            "[",
            "1",
            "3",
            ";",
            "2",
            "u",
            key_str=("s-enter",),
            group="Notebook",
            desc="Run then select next cell",
        )
        @kb.add("c-r", group="Notebook", desc="Run then select next cell")
        @kb.add("f21")
        def run_then_next(event: "KeyPressEvent") -> "None":
            # Insert a cell if we are at the last cell
            n_cells = len(self.nb.page.children)
            if self.nb.page.selected_index == (n_cells) - 1:
                offset = n_cells - self.nb.page.selected_index
                self.nb.add(offset)
            else:
                self.nb.page.selected_index += 1
            run_or_render(event)

        @kb.add("c-f", filter=self.is_editing, group="Edit Mode", desc="Find")
        def find(event: "KeyPressEvent") -> "None":
            start_search(self.input_box.control)

        @kb.add("c-g", filter=self.is_editing, group="Edit Mode", desc="Find Next")
        def find_next(event: "KeyPressEvent") -> "None":
            search_state = get_app().current_search_state
            cursor_position = self.input_box.buffer.get_search_position(
                search_state, include_current_position=False
            )
            self.input_box.buffer.cursor_position = cursor_position

        @kb.add("c-z", filter=self.is_editing, group="Edit Mode", desc="Undo")
        def undo(event: "KeyPressEvent") -> "None":
            self.input_box.buffer.undo()

        @kb.add("c-d", filter=self.is_editing, group="Edit Mode", desc="Duplicate line")
        def duplicate_line(event: "KeyPressEvent") -> "None":
            buffer = event.current_buffer
            line = buffer.document.current_line
            eol = buffer.document.get_end_of_line_position()
            buffer.cursor_position += eol
            buffer.newline()
            buffer.insert_text(line)
            buffer.cursor_position -= eol

        @kb.add("home", filter=self.is_editing)
        def smart_home(event: "KeyPressEvent") -> "None":
            buffer = event.current_buffer
            buffer.cursor_position += buffer.document.get_start_of_line_position(
                after_whitespace=buffer.document.get_start_of_line_position(
                    after_whitespace=True
                )
                != 0
            )

        @kb.add("enter", filter=self.is_editing)
        def new_line(event: "KeyPressEvent") -> "None":
            buffer = event.current_buffer
            buffer.cut_selection()
            pre = buffer.document.text_before_cursor
            buffer.newline()
            if pre.rstrip()[-1:] in (":", "(", "["):
                dent_buffer(event)

        def dent_buffer(event: "KeyPressEvent", un: "bool" = False) -> "None":
            buffer = event.current_buffer
            selection_state = buffer.selection_state
            cursor_position = buffer.cursor_position
            lines = buffer.document.lines

            # Apply indentation to the selected range
            from_, to = map(
                lambda x: buffer.document.translate_index_to_position(x)[0],
                buffer.document.selection_range(),
            )
            dent = unindent if un else indent
            dent(buffer, from_, to + 1, count=event.arg)

            # If there is a selection, indent it and adjust the selection range
            if selection_state:
                change = 4 * (un * -2 + 1)
                # Count how many lines will be affected
                line_count = 0
                for i in range(from_, to + 1):
                    if not un or lines[i][:1] == " ":
                        line_count += 1
                backwards = cursor_position < selection_state.original_cursor_position
                if un and not line_count:
                    buffer.cursor_position = cursor_position
                else:
                    buffer.cursor_position = max(
                        0, cursor_position + change * (1 if backwards else line_count)
                    )
                    selection_state.original_cursor_position = max(
                        0,
                        selection_state.original_cursor_position
                        + change * (line_count if backwards else 1),
                    )

            # Maintain the selection state before indentation
            buffer.selection_state = selection_state

        @kb.add(
            "tab",
            filter=self.is_editing & (cursor_in_leading_ws | has_selection),
            group="Edit Mode",
            desc="Indent",
        )
        def indent_buffer(event: "KeyPressEvent") -> "None":
            dent_buffer(event)

        @kb.add(
            "s-tab",
            filter=self.is_editing & (cursor_in_leading_ws | has_selection),
            group="Edit Mode",
            desc="Unindent",
        )
        def unindent_buffer(event: "KeyPressEvent") -> "None":
            dent_buffer(event, un=True)

        @kb.add("escape", filter=has_completions, eager=True)
        def cancel_completion(event: "KeyPressEvent") -> "None":
            """Cancel a completion with the escape key."""
            event.current_buffer.cancel_completion()

        @kb.add("enter", filter=has_completions)
        def apply_completion(event: "KeyPressEvent") -> "None":
            """Cancel a completion with the escape key."""
            complete_state = event.current_buffer.complete_state
            if complete_state:
                assert isinstance(complete_state.current_completion, Completion)
                event.current_buffer.apply_completion(complete_state.current_completion)

        @kb.add("c-c", filter=self.is_editing, group="Edit Mode", desc="Copy")
        def copy_selection(event: "KeyPressEvent") -> "None":
            data = event.current_buffer.copy_selection()
            get_app().clipboard.set_data(data)

        @kb.add(
            "c-x", filter=self.is_editing, eager=True, group="Edit Mode", desc="Cut"
        )
        def cut_selection(event: "KeyPressEvent") -> "None":
            data = event.current_buffer.cut_selection()
            get_app().clipboard.set_data(data)

        @kb.add("c-v", filter=self.is_editing, group="Edit Mode", desc="Paste")
        def paste_clipboard(event: "KeyPressEvent") -> "None":
            event.current_buffer.paste_clipboard_data(get_app().clipboard.get_data())

        return kb

    def run(self) -> "None":
        """Run the contents of a code cell in the kernel."""
        self.clear_output()
        if self.nb.kc and self.nb.kernel_loop is not None:
            # Execute input and wait for responses in kernel thread
            asyncio.run_coroutine_threadsafe(
                # self.nb.kc._async_execute_interactive(
                self._async_execute_interactive(
                    code=self.input,
                    allow_stdin=False,
                    output_hook=self.ran,
                ),
                self.nb.kernel_loop,
            )

    async def _async_execute_interactive(
        self,
        code: "str",
        output_hook: "Callable[[dict[str, Any]], None]",
        allow_stdin: "bool" = False,
    ) -> "dict[str, Any]":
        from queue import Empty

        import zmq.asyncio

        assert self.nb.kc is not None
        if not self.nb.kc.iopub_channel.is_alive():
            raise RuntimeError("IOPub channel must be running to receive output")

        msg_id = self.nb.kc.execute(
            code,
            allow_stdin=False,
        )
        stdin_hook = self.nb.kc._stdin_hook_default

        timeout_ms = None

        assert hasattr(zmq, "Poller")
        assert hasattr(zmq, "POLLIN")

        poller = zmq.Poller()
        iopub_socket = self.nb.kc.iopub_channel.socket
        poller.register(iopub_socket, zmq.POLLIN)
        stdin_socket = None

        # wait for output and redisplay it
        while True:
            events = dict(poller.poll(timeout_ms))
            if not events:
                raise TimeoutError("Timeout waiting for output")
            if stdin_socket in events:
                req = self.nb.kc.stdin_channel.get_msg(timeout=0)
                stdin_hook(req)
                continue
            if iopub_socket not in events:
                continue

            msg = self.nb.kc.iopub_channel.get_msg(timeout=0)

            if msg["parent_header"].get("msg_id") != msg_id:
                # not from my request
                continue
            output_hook(msg)

            # stop on idle
            if (
                msg["header"]["msg_type"] == "status"
                and msg["content"]["execution_state"] == "idle"
            ):
                break

        # output is done, get the reply
        while True:
            try:
                reply = self.nb.kc.get_shell_msg(timeout=None)
            except Empty as e:
                raise TimeoutError("Timeout waiting for reply") from e
            if reply["parent_header"].get("msg_id") != msg_id:
                # not my reply, someone may have forgotten to retrieve theirs
                continue
            return reply

    def ran(self, msg: "dict[str, Any]") -> "None":
        """Callback which runs when a message for this cell is recieved from the kernel."""
        msg_type = msg.get("header", {}).get("msg_type")

        if msg_type == "status":
            self.state = msg.get("content", {}).get("execution_state")
            self.nb.kernel_status = self.state

        elif msg_type == "execute_input":
            self.execution_count = msg.get("content", {}).get("execution_count")

        elif msg_type == "stream":
            name = msg.get("content", {}).get("name")
            outputs = {
                output.get("name"): output
                for output in self.outputs
                if output.get("name")
            }
            if name in outputs:
                outputs[name]["text"] += msg.get("content", {}).get("text", "")
            else:
                self.add_output(nbformat.v4.output_from_msg(msg))

        elif msg_type in ("error", "display_data", "execute_result"):
            self.add_output(nbformat.v4.output_from_msg(msg))

        # Update the outputs in the visible instance of this cell
        visible_cell = self.nb.get_cell_by_id(self.id)
        if visible_cell:
            visible_cell.output_box.children = visible_cell.rendered_outputs

        # Tell the app that the display needs updating
        get_app().invalidate()

    def set_cell_type(self, cell_type: "Literal['markdown','code','raw']") -> "None":
        """Convert the cell to a different cell type.

        Args:
            cell_type: The desired cell type.

        """
        if cell_type == "code":
            self.json.setdefault("execution_count", None)
        self.json["cell_type"] = cell_type
        self.load()

    def load(self) -> "None":
        """Generates the main container used to represent a notebook cell."""
        fill = partial(Window, style=self.border_style)

        self.search_control = SearchToolbar()

        self.input_box = TextArea(
            text=self.input,
            # Does not accept conditions
            scrollbar=self.scroll_input(),
            wrap_lines=self.wrap_input,
            # Does not accept conditions
            line_numbers=self.show_input_line_numbers(),
            read_only=~self.is_editing,
            focusable=self.is_editing,
            lexer=DynamicLexer(
                lambda: PygmentsLexer(
                    get_lexer_by_name(self.language).__class__,
                    sync_from_start=False,
                )
                if self.cell_type != "raw"
                else None
            ),
            search_field=self.search_control,
            completer=self.nb.completer,
            complete_while_typing=False,
            style="class:cell-input",
        )
        self.input_box.window.cursorline = self.is_editing
        self.input_box.buffer.tempfile_suffix = ".py"

        self.output_box = HSplit(
            self.rendered_outputs,
            style="class:cell-output",
        )

        top_border = VSplit(
            [
                self.control,
                ConditionalContainer(
                    content=fill(
                        char=Border.HORIZONTAL, width=lambda: len(self.prompt), height=1
                    ),
                    filter=self.show_prompt,
                ),
                ConditionalContainer(
                    content=fill(width=1, height=1, char=Border.SPLIT_TOP),
                    filter=self.show_prompt,
                ),
                fill(char=Border.HORIZONTAL, height=1),
                fill(width=1, height=1, char=Border.TOP_RIGHT),
            ],
            height=1,
        )
        input_row = ConditionalContainer(
            VSplit(
                [
                    fill(width=1, char=Border.VERTICAL),
                    ConditionalContainer(
                        content=Window(
                            FormattedTextControl(
                                lambda: self.prompt,
                            ),
                            width=lambda: len(self.prompt),
                            style="class:cell-input-prompt",
                        ),
                        filter=self.show_prompt,
                    ),
                    ConditionalContainer(
                        content=fill(width=1, char=Border.VERTICAL),
                        filter=self.show_prompt,
                    ),
                    HSplit([self.input_box, self.search_control]),
                    fill(width=1, char=Border.VERTICAL),
                ],
            ),
            filter=self.show_input,
        )
        middle_line = ConditionalContainer(
            content=VSplit(
                [
                    fill(width=1, height=1, char=Border.SPLIT_LEFT),
                    ConditionalContainer(
                        content=fill(
                            char=Border.HORIZONTAL, width=lambda: len(self.prompt)
                        ),
                        filter=self.show_prompt,
                    ),
                    ConditionalContainer(
                        content=fill(width=1, height=1, char=Border.CROSS),
                        filter=self.show_prompt,
                    ),
                    fill(char=Border.HORIZONTAL),
                    fill(width=1, height=1, char=Border.SPLIT_RIGHT),
                ],
                height=1,
            ),
            filter=self.show_input & self.show_output,
        )
        output_row = ConditionalContainer(
            VSplit(
                [
                    fill(width=1, char=Border.VERTICAL),
                    ConditionalContainer(
                        content=Window(
                            FormattedTextControl(
                                lambda: self.prompt,
                            ),
                            width=lambda: len(self.prompt),
                            style="class:cell-output-prompt",
                        ),
                        filter=self.show_prompt,
                    ),
                    ConditionalContainer(
                        fill(width=1, char=" "), filter=~self.show_prompt
                    ),
                    ConditionalContainer(
                        content=fill(width=1, char=Border.VERTICAL),
                        filter=self.show_prompt,
                    ),
                    self.output_box,
                    ConditionalContainer(
                        fill(width=1, char=" "), filter=~self.show_prompt
                    ),
                    fill(width=1, char=Border.VERTICAL),
                ],
            ),
            filter=self.show_output,
        )
        bottom_border = VSplit(
            [
                fill(width=1, height=1, char=Border.BOTTOM_LEFT),
                ConditionalContainer(
                    content=fill(
                        char=Border.HORIZONTAL, width=lambda: len(self.prompt)
                    ),
                    filter=self.show_prompt,
                ),
                ConditionalContainer(
                    content=fill(width=1, height=1, char=Border.SPLIT_BOTTOM),
                    filter=self.show_prompt,
                ),
                fill(char=Border.HORIZONTAL),
                fill(width=1, height=1, char=Border.BOTTOM_RIGHT),
            ],
            height=1,
        )

        self.container = FloatContainer(
            content=HSplit(
                [top_border, input_row, middle_line, output_row, bottom_border],
                key_bindings=self.load_key_bindings(),
            ),
            floats=[
                Float(
                    transparent=True,
                    left=0,
                    right=0,
                    top=0,
                    bottom=0,
                    content=ConditionalContainer(
                        ClickArea(self), filter=~self.is_focused
                    ),
                ),
            ],
        )

    def border_style(self) -> "str":
        """Determines the style of the cell borders, based on the cell state."""
        if self.focused:
            if self.editing:
                return "class:frame.border,cell-border-edit"
            else:
                return "class:frame.border,cell-border-selected"
        else:
            return "class:frame.border,cell-border"

    @property
    def id(self) -> "str":
        """Returns the cell's ID as per the cell JSON."""
        return self.json.get("id", "")

    @property
    def language(self) -> "str":
        """Returns the cell's code language."""
        if self.cell_type == "markdown":
            return "markdown"
        else:
            return self.nb.json.metadata.get("language_info", {}).get("name", "python")

    @property
    def focused(self) -> "bool":
        """Determine if the cell currently has focus."""
        return get_app().layout.has_focus(self.container)

    @property
    def cell_type(self) -> "str":
        """Determine the currrent cell type."""
        return self.json.get("cell_type", "code")

    @property
    def prompt(self) -> "str":
        """Determine what should be displayed in the prompt of the cell."""
        if self.state in ("busy", "queued"):
            prompt = "*"
        else:
            prompt = self.execution_count
        if prompt is None:
            prompt = " "
        if prompt:
            prompt = f"[{prompt}]"
        return prompt

    @property
    def execution_count(self) -> "str":
        """Retrieve the execution count from the cell's JSON."""
        return self.json.get("execution_count", " ")

    @execution_count.setter
    def execution_count(self, count: int) -> "None":
        """Set the execution count in the cell's JSON.

        Args:
            count: The new execution count number.

        """
        self.json["execution_count"] = count

    @property
    def input(self) -> "str":
        """Fetch the cell's contents from the cell's JSON."""
        return self.json.get("source", "")

    @input.setter
    def input(self, value: "str") -> "None":
        """Set the cell's contents in the cell's JSON.

        Args:
            value: The new cell contents text.

        """
        self.json["source"] = value

    def clear_output(self) -> "None":
        """Remove all outputs from the cell."""
        self.json["outputs"] = []
        self.load()

    @property
    def outputs(self) -> "list[dict[str, Any]]":
        """Retrieve a list of cell outputs from the cell's JSON."""
        if self.cell_type == "markdown":
            return [
                {"data": {"text/x-markdown": self.input}, "output_type": "markdown"}
            ]
        else:
            return self.json.get("outputs", [])

    def add_output(self, output: "dict[str, Any]") -> "None":
        """Append a new output to the cell's JSON.

        Args:
            output: The output JSON to add.

        """
        self.json.setdefault("outputs", []).append(output)
        self.output_box.children = self.rendered_outputs

    @property
    def rendered_outputs(self) -> "list[Container]":
        """Generates a list of rendered outputs."""
        rendered_outputs: "list[Container]" = []
        for i, output_json in enumerate(self.outputs):
            rendered_outputs.append(to_container(Output(i, output_json, parent=self)))
        return rendered_outputs

    def __pt_container__(self) -> "Container":
        """Returns the container which represents this cell."""
        return self.container



# -*- coding: utf-8 -*-
"""Contains the `ScrollingContainer` class, which renders children on the fly."""
from __future__ import annotations

from collections import namedtuple
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Sequence, Union, cast

from prompt_toolkit.application.current import get_app
from prompt_toolkit.data_structures import Point
from prompt_toolkit.filters import FilterOrBool, to_filter
from prompt_toolkit.layout.containers import (
    AnyContainer,
    Container,
    Window,
    to_container,
)
from prompt_toolkit.layout.dimension import (
    AnyDimension,
    Dimension,
    max_layout_dimensions,
    to_dimension,
)
from prompt_toolkit.layout.mouse_handlers import MouseHandler, MouseHandlers
from prompt_toolkit.layout.screen import Char, Screen, WritePosition
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from prompt_toolkit.utils import to_str

from euporie.box import Border
from euporie.config import config
from euporie.keys import KeyBindingsInfo

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindings
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent

    from euporie.app import App

DrawingPosition = namedtuple(
    "DrawingPosition",
    [
        "index",
        "container",
        "top",
        "height",
        "parent_height",
    ],
)


class ScrollingContainer(Container):
    """A scrollable container which renders only the currently visible children."""

    def __init__(
        self,
        children: "list[Union[Callable, AnyContainer]]",
        max_content_width: "AnyDimension" = None,
        height: "AnyDimension" = None,
        z_index: "Optional[int]" = None,
        style: "Union[str, Callable[[], str]]" = "",
        show_border: "FilterOrBool" = True,
    ):
        """Initiates the `ScrollingContainer`."""
        self.children = children
        if not isinstance(max_content_width, Dimension):
            max_content_width = Dimension(preferred=int(config.max_notebook_width))
        self.max_content_width = max_content_width
        self.height = height
        self.z_index = z_index
        self.style = style

        self.key_bindings = self.load_key_bindings()

        self._remaining_space_window = Window()

        self.show_border = to_filter(show_border)

        self.last_selected_index: "int" = 0

        # Position of viewing window relative to selected child
        self.selected_child_position: "int" = 0
        self.child_cache: "dict[int, AnyContainer]" = {}
        self.size_cache: "dict[int, int]" = {}
        self.to_draw: "Sequence[DrawingPosition]" = []

    # Container methods

    def reset(self) -> "None":
        """Reset the state of rendered children."""
        for c in self.get_children():
            c.reset()
        self.child_cache = {}
        self.size_cache = {}

    def preferred_width(self, max_available_width: "int") -> "Dimension":
        """Do not provide a preferred width - grow to fill the avaliable space."""
        return Dimension()

    def preferred_height(self, width: int, max_available_height: int) -> "Dimension":
        """Return the preferred height only if one is provided."""
        if self.height is not None:
            return to_dimension(self.height)
        return Dimension()  # min=max_available_height, max=max_available_height)

    def write_to_screen(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        write_position: "WritePosition",
        parent_style: "str",
        erase_bg: "bool",
        z_index: "Optional[int]",
    ) -> None:
        """Render the container to a `Screen` instance.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        self.available_width = write_position.width
        self.available_height = write_position.height

        style = parent_style + " " + to_str(self.style)
        z_index = z_index if self.z_index is None else self.z_index

        self.content_width = min(
            self.max_content_width.preferred
            if self.max_content_width.preferred
            else self.available_width,
            write_position.width - self.show_scroll * 2,
        )
        self.content_height = self.available_height

        self.to_draw = self.arrange_children(write_position)

        ypos = write_position.ypos
        xpos = (
            write_position.xpos
            + (write_position.width - self.show_scroll * 2 - self.content_width) // 2
        )

        # Draw frame if it fits
        if (
            self.show_border()
            and self.content_width + 4 + self.show_scroll <= self.available_width
        ):
            for y in range(ypos, ypos + self.content_height):
                screen.data_buffer[y][xpos - 2] = Char(Border.VERTICAL)
                screen.data_buffer[y][xpos + 1 + self.content_width] = Char(
                    Border.VERTICAL
                )

        # Draw background if there is space
        if config.background_pattern:
            dot = Char(str(config.background_character), "class:background")
            for y in range(ypos, ypos + self.available_height):
                for xrange in (
                    (write_position.xpos, xpos - 2),
                    (xpos + self.content_width + 2, write_position.width - 1),
                ):
                    for x in range(*xrange):
                        if (
                            (config.background_pattern == 1 and (x + y) % 2 == 0)
                            or (config.background_pattern == 2 and (x + 2 * y) % 4 == 0)
                            or (config.background_pattern == 3 and (x + y) % 3 == 0)
                            or (
                                config.background_pattern == 4
                                and ((x + y % 2 * 3) % 6) % 4 == 0
                            )
                        ):
                            screen.data_buffer[y][x] = dot

        # Draw child panes.
        for drawing in self.to_draw:

            if drawing.top < 0 or drawing.top + drawing.height > self.content_height:
                # Create a virtual screen to draw partially obscured children
                temp_screen = Screen(default_char=Char(char=" ", style=parent_style))
                temp_screen.show_cursor = False
                # Create empty mouse handler array
                temp_mouse_handlers = MouseHandlers()
                # Draw child at (0,0) on temp screen
                temp_write_position = WritePosition(
                    xpos=0, ypos=0, width=self.content_width, height=drawing.height
                )

                drawing.container.write_to_screen(
                    temp_screen,
                    temp_mouse_handlers,
                    temp_write_position,
                    style,
                    erase_bg,
                    z_index,
                )
                temp_screen.draw_all_floats()

                # Determin how many rows to copy
                if drawing.top < 0:
                    copy_rows = drawing.top + drawing.height
                else:  # drawing.top + drawing.height > self.content_height
                    copy_rows = self.content_height - drawing.top

                # Copy screen contents
                for i in range(0, copy_rows):

                    real_y = ypos + max(0, drawing.top) + i
                    virt_y = 0 - min(0, drawing.top) + i

                    real_row = screen.data_buffer[real_y]
                    virt_row = temp_screen.data_buffer[virt_y]

                    real_mouse_row = mouse_handlers.mouse_handlers[real_y]
                    virt_mouse_row = temp_mouse_handlers.mouse_handlers[virt_y]

                    # real_escapes = screen.zero_width_escapes[real_y]
                    # virt_escapes = temp_screen.zero_width_escapes[virt_y]

                    for x in range(self.content_width):
                        real_row[x + xpos] = virt_row[x]
                        real_mouse_row[x + xpos] = virt_mouse_row[x]
                        # real_escapes[x + xpos] = virt_escapes[x]

                # Copy cursors
                if temp_screen.show_cursor:
                    screen.show_cursor = True
                for window, point in temp_screen.cursor_positions.items():
                    # Check cursor is in the visibible zone on the temp screen
                    if (
                        0 <= point.x < self.content_width
                        and 0 - min(0, drawing.top)
                        <= point.y
                        < 0 - min(0, drawing.top) + copy_rows
                    ):
                        screen.cursor_positions[window] = Point(
                            x=point.x + xpos,
                            y=point.y + ypos + drawing.top,
                        )

            # If the cell is fully visible, simply write it to the screen
            else:
                drawing.container.write_to_screen(
                    screen,
                    mouse_handlers,
                    WritePosition(
                        xpos, ypos + drawing.top, self.content_width, drawing.height
                    ),
                    style,
                    erase_bg,
                    z_index,
                )

        # TODO - Fill empty space, and click to focus

        # Draw floats before wrapping mouse handlers
        screen.draw_all_floats()

        # Draw scroll bar
        if self.show_scroll:

            x = write_position.xpos + self.available_width - 1

            def _mouse_handler_scroll_up(mouse_event: MouseEvent) -> None:
                if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
                    self.scroll_up()

            def _mouse_handler_scroll_down(mouse_event: MouseEvent) -> None:
                if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
                    self.scroll_down()

            screen.data_buffer[ypos + 0][x] = Char("▲", "class:scrollbar.arrow")
            mouse_handlers.mouse_handlers[ypos][x] = _mouse_handler_scroll_up
            screen.data_buffer[ypos + self.available_height - 1][x] = Char(
                "▼", "class:scrollbar.arrow"
            )
            mouse_handlers.mouse_handlers[ypos + self.available_height - 1][
                x
            ] = _mouse_handler_scroll_down

            def _scrollbar_mouse_click(index: "int") -> "Callable[[MouseEvent], None]":
                def mouse_handler(mouse_event: MouseEvent) -> "None":
                    if mouse_event.event_type == MouseEventType.MOUSE_DOWN:
                        self.selected_index = index

                return mouse_handler

            scroll_tab_height = max(
                1, (self.available_height - 2) // len(self.children)
            )
            scroll_tab_top = (
                (self.available_height - 2 - scroll_tab_height)
                * self.to_draw[0].index
                // (len(self.children) - 1)
                if self.to_draw
                else 0
            )

            for y in range(self.available_height - 2):
                if scroll_tab_top <= y <= scroll_tab_height + scroll_tab_top:
                    style = "class:scrollbar.button"
                else:
                    style = "class:scrollbar.background"
                    # Set mouse click action
                child_index = int(
                    y / (self.available_height - 2 - 1) * len(self.children)
                )
                mouse_handlers.mouse_handlers[ypos + y + 1][x] = _scrollbar_mouse_click(
                    child_index
                )
                screen.data_buffer[ypos + y + 1][x] = Char(style=style)

        # Wrap all the mouse handlers to add mouse scrolling
        mouse_handler_wrappers: Dict[MouseHandler, MouseHandler] = {}

        def _wrap_mouse_handler(handler: "Callable") -> "Callable[[MouseEvent], None]":
            if handler not in mouse_handler_wrappers:

                def wrapped_mouse_handler(mouse_event: MouseEvent) -> "None":
                    if mouse_event.event_type == MouseEventType.SCROLL_DOWN:
                        self.scroll_down()
                    elif mouse_event.event_type == MouseEventType.SCROLL_UP:
                        self.scroll_up()
                    elif handler:
                        handler(mouse_event)

                mouse_handler_wrappers[handler] = wrapped_mouse_handler
            return mouse_handler_wrappers[handler]

        for y in range(ypos, ypos + write_position.height):
            mouse_row = mouse_handlers.mouse_handlers[y]
            for x in range(
                write_position.xpos, write_position.xpos + write_position.width
            ):
                mouse_row[x] = _wrap_mouse_handler(mouse_row[x])

        # If we are using kitty graphics, ensure all existing images are erased
        if cast("App", get_app()).has_kitty_graphics:
            screen.zero_width_escapes[ypos][xpos] = "\x1b_Ga=d\x1b\\"

    @property
    def show_scroll(self) -> "bool":
        """Determine whether the scrollbar show be shown."""
        return len(self.children) > 1
        # return sum([drawing.height for drawing in self.to_draw]) > self.content_height

    def get_children(self) -> "List[Container]":
        """Return a list of the containers of the currently rendered children."""
        return list(map(to_container, [x[1] for x in sorted(self.child_cache.items())]))

    def get_key_bindings(self) -> "KeyBindings":
        """Return the keybindings for the `ScrollingContainer`."""
        return self.key_bindings

    # End of container methods

    def load_key_bindings(self) -> "KeyBindings":
        """Loads the keybindings for the `ScrollingContainer`."""
        kb = KeyBindingsInfo()

        @kb.add("[", group="Navigation", desc="Scroll up")
        @kb.add("<scroll-up>")
        def su(event: "KeyPressEvent") -> None:
            self.scroll_up()

        @kb.add("]", group="Navigation", desc="Scroll down")
        @kb.add("<scroll-down>")
        def sd(event: "KeyPressEvent") -> None:
            self.scroll_down()

        @kb.add("c-up", group="Navigation", desc="Go to first cell")
        @kb.add("home", group="Navigation", desc="Go to first cell")
        def first_child(event: "KeyPressEvent") -> None:
            self.selected_index = 0

        @kb.add("pageup", group="Navigation", desc="Go up 5 cells")
        def prev_child5(event: "KeyPressEvent") -> None:
            self.selected_index -= 5

        @kb.add("up", group="Navigation", desc="Go up one cell")
        @kb.add("k", group="Navigation", desc="Go up one cell")
        def prev_child(event: "KeyPressEvent") -> None:
            self.selected_index -= 1

        @kb.add("down", group="Navigation", desc="Go down one cell")
        @kb.add("j", group="Navigation", desc="Go down one cell")
        def next_child(event: "KeyPressEvent") -> None:
            self.selected_index += 1

        @kb.add("pagedown", group="Navigation", desc="Go down 5 cells")
        def next_child5(event: "KeyPressEvent") -> None:
            self.selected_index += 5

        @kb.add("c-down", group="Navigation", desc="Go to last cell")
        @kb.add("end", group="Navigation", desc="Go to last cell")
        def last_child(event: "KeyPressEvent") -> None:
            self.selected_index = len(self.children)

        return kb

    def get_child(self, index: "Optional[int]" = None) -> "AnyContainer":
        """Return a rendered instance of the child at the given index.

        If no index is given, the currently selected child is returned.

        Args:
            index: The index of the child to return.

        Returns:
            A rendered instance of the child.

        """
        if index is None:
            index = self.last_selected_index
        child = self.child_cache.get(index)
        if child is None and self.children:
            container_or_generator = self.children[index]
            if callable(container_or_generator):
                child = container_or_generator()
            else:
                child = container_or_generator
            self.child_cache[index] = child
        assert child is not None
        return child

    def get_child_size(self, index: "int", refresh: "bool" = False) -> "int":
        """Return the last known height of the child at the given index.

        For unrendered children, the height is retrieved from a cache unless `refresh`
        is `True`.

        Args:
            index: The index of the child for which to get the last known height.
            refresh: Whether to render an underendered child to calculate its height.

        Returns:
            The height of the child in rows.

        """
        if not refresh and (size := self.size_cache.get(index)):
            return size
        else:
            child = self.get_child(index)
            # assert isinstance(child, AnyContainer)
            container = to_container(child)
            size = container.preferred_height(
                self.content_width, self.content_height
            ).preferred
            self.size_cache[index] = size
            return size

    def arrange_children(
        self, write_position: "WritePosition"
    ) -> "List[DrawingPosition]":
        """Calculate which children should be draw and at which positions.

        Args:
            write_position: Defines the space available for drawing children.

        Returns:
            A list of `DrawingPosition`, detailing the positions at which the children
                should be drawn to the screen.

        """
        selected_index = self.selected_index

        # Create a new version of the child cache
        cache = {}

        # Get a handle on the current child
        # child = self.get_child(selected_index)
        # container = to_container(child)
        # cache[selected_index] = child
        # size = self.get_child_size(selected_index, refresh=True)

        # Empy the list of container
        self.to_draw = []

        # Display selected child and those below it that are on screen
        if self.selected_child_position <= self.content_height:
            top = self.selected_child_position
            for i in range(self.selected_index, len(self.children)):
                size = self.get_child_size(i)
                if 0 <= top + size:
                    cache[i] = self.get_child(i)
                    container = to_container(cache[i])
                    size = self.get_child_size(i, refresh=True)
                    drawing_position = DrawingPosition(
                        index=i,
                        container=container,
                        top=top,
                        height=size,
                        parent_height=self.content_height,
                    )
                    self.to_draw.append(drawing_position)
                top += size
                if top >= self.content_height:
                    break

        # Add children to fill space above the selected child if any are on screen
        if self.selected_child_position > 0:
            top = self.selected_child_position
            for i in range(selected_index - 1, -1, -1):
                size = self.get_child_size(i)
                if top >= 0:
                    cache[i] = self.get_child(i)
                    container = to_container(cache[i])
                    size = self.get_child_size(i, refresh=True)
                    drawing_position = DrawingPosition(
                        index=i,
                        container=container,
                        top=top - size,
                        height=size,
                        parent_height=self.content_height,
                    )
                    self.to_draw.append(drawing_position)
                top -= size
                if top <= 0:
                    break

        # Add additional children to cache
        if cache:
            next_child_index = min(len(self.children) - 1, max(cache.keys()) + 1)
            if next_child_index not in cache:
                cache[next_child_index] = self.get_child(next_child_index)
            prev_child_index = max(0, min(cache.keys()) - 1)
            if prev_child_index not in cache:
                cache[prev_child_index] = self.get_child(prev_child_index)
        # Ensure the selected child is in the cache
        if self.selected_index not in cache:
            cache[self.selected_index] = self.get_child(self.selected_index)
        # Replace the cache
        self.child_cache = cache

        # Sort drawings
        self.to_draw = sorted(self.to_draw, key=lambda x: x.index)

        return self.to_draw

    def is_child_obscured(self, index: "int") -> "bool":
        """Determine if a child is partially visible.

        Args:
            index: The index of the child of interest.

        Returns:
            True if the child is rendered and partially off-screen, otherwise False.=

        """
        drawing = {drawing.index: drawing for drawing in self.to_draw}.get(index)
        return drawing is None or (
            drawing.top < 0 or drawing.parent_height < drawing.top + drawing.height
        )

    def select_child(self, new_index: "int") -> "None":
        """Focus a child and scroll so it is visible.

        Args:
            new_index: The index of the child to select.

        """
        # Get the height and position of the newly selected child
        # Check if the newly selected child is currently in view
        drawing = {drawing.index: drawing for drawing in self.to_draw}.get(new_index)
        # If it is displayed on the screen, we already have that information
        if drawing:
            new_child_height = drawing.height
            new_top = drawing.top
        # If it is not displayed on the screen, we can load it from the child cache or
        # render it to calculate it's size. If it's not drawn, we pretend its position
        # is off the top of the screen or below the bottom, depending on if it is above
        # or below the currently selected child.
        else:
            new_child_height = self.get_child_size(
                new_index,
            )
            # If the children do not overflow the screen, move to the position of the
            # lowest child instead of the bottom of the screen
            new_top = (
                min(
                    max([drawing.top + drawing.height for drawing in self.to_draw]),
                    self.content_height,
                )
                if new_index > self.last_selected_index
                else 0
            )

        # If the newly selected is off the top of the display, position it at the top
        if new_top < 0:
            self.selected_child_position = 0
        # If it is off the bottom of the screen, move it fully into view at the bottom
        elif new_top > self.content_height - new_child_height:
            self.selected_child_position = max(
                0, self.content_height - new_child_height
            )
        # Otherwise, it is fully on screen, so move the selection to it's position
        else:
            self.selected_child_position = new_top

    def scroll_up(self) -> None:
        """Scrolls up 1 row."""
        for drawing in self.to_draw:
            if drawing.index == 0:
                if drawing.top + 1 > 0:
                    return
                break
        self.selected_child_position += 1

    def scroll_down(self) -> None:
        """Scrolls down 1 row."""
        for drawing in self.to_draw:
            if drawing.index == len(self.children) - 1:
                if drawing.top + drawing.height < self.content_height // 2:
                    return
                break
        self.selected_child_position -= 1

    @property
    def show_scrollbar(self) -> "bool":
        """Determines if the scrollbar should be shown."""
        return sum([x.height for x in self.to_draw]) >= self.content_height

    @property
    def selected_index(self) -> "int":
        """Returns in index of the currently selected child."""
        app = get_app()
        # Detect if focused child element has changed
        if app.layout.has_focus(self):
            # Find index of selected child
            index = self.last_selected_index
            for i, child in self.child_cache.items():
                if app.layout.has_focus(child):
                    index = i
                    break
            # This will change the position when a new child is selected
            self.selected_index = index
        return self.last_selected_index

    @selected_index.setter
    def selected_index(self, new_index: "int") -> "None":
        """Sets the currently selected child index.

        Args:
            new_index: The index of the child to select.

        """
        self._set_selected_index(new_index)

    def _set_selected_index(self, new_index: "int", force: "bool" = False) -> None:
        # Only update the selected child if it was not selected before
        if force or new_index != self.last_selected_index:
            # Ensure selected index is a valid child
            new_index = min(max(new_index, 0), len(self.children) - 1)
            self.select_child(new_index)

            # Focus new child if not already focused
            self.focus(new_index)

            # Track which child was selected
            self.last_selected_index = new_index

    def focus(self, index: "Optional[int]" = None) -> "None":
        """Focuses the child with the given index.

        Args:
            index: The index of the child to focus.
        """
        if index is None:
            index = self.last_selected_index
        child = self.get_child(index)
        app = get_app()
        if not app.layout.has_focus(child):
            app.layout.focus(child)


class PrintingContainer(Container):
    """A container which displays all it's children in a vertical list."""

    def __init__(
        self,
        children: "list[Union[Callable, AnyContainer]]",
        width: "AnyDimension" = None,
    ):
        """Initiate the container."""
        self.width = width
        self.renderers = children
        self.rendered = False
        self._children: "list[Container]" = []

    def write_to_screen(
        self,
        screen: "Screen",
        mouse_handlers: "MouseHandlers",
        write_position: "WritePosition",
        parent_style: "str",
        erase_bg: "bool",
        z_index: "Optional[int]",
    ) -> "None":
        """Render the container to a `Screen` instance.

        All children are rendered vertically in sequence.

        Args:
            screen: The :class:`~prompt_toolkit.layout.screen.Screen` class to which
                the output has to be written.
            mouse_handlers: :class:`prompt_toolkit.layout.mouse_handlers.MouseHandlers`.
            write_position: A :class:`prompt_toolkit.layout.screen.WritePosition` object
                defining where this container should be drawn.
            erase_bg: If true, the background will be erased prior to drawing.
            parent_style: Style string to pass to the :class:`.Window` object. This will
                be applied to all content of the windows. :class:`.VSplit` and
                :class:`prompt_toolkit.layout.containers.HSplit` can use it to pass
                their style down to the windows that they contain.
            z_index: Used for propagating z_index from parent to child.

        """
        xpos = write_position.xpos
        ypos = write_position.ypos

        for child in self._all_children:
            height = child.preferred_height(write_position.width, 99999).preferred
            child.write_to_screen(
                screen,
                mouse_handlers,
                WritePosition(xpos, ypos, write_position.width, height),
                parent_style,
                erase_bg,
                z_index,
            )
            ypos += height

    @property
    def _all_children(self) -> "list[Container]":
        """Calculates and returns a list of all child containers.

        If some children are generator functions, children will be generated first.

        Returns:
            A list of child containers

        """
        if not self.rendered:
            for renderer in self.renderers:
                if callable(renderer):
                    self._children.append(to_container(renderer()))
                else:
                    self._children.append(to_container(renderer))
            self.rendered = True
        return self._children

    def get_children(self) -> "list[Container]":
        """Returns a list of all child containers."""
        return self._children

    def preferred_height(self, width: int, max_available_height: int) -> "Dimension":
        """Returns the preferred height, equal to the sum of the child heights."""
        return Dimension(
            min=1,
            preferred=sum(
                [
                    c.preferred_height(width, max_available_height).preferred
                    for c in self._all_children
                ]
            ),
        )

    def preferred_width(self, max_available_width: "int") -> "Dimension":
        """Calculates and returns the desired width for this container."""
        if self.width is not None:
            dim = to_dimension(self.width).preferred
            return Dimension(max=dim, preferred=dim)

        if self._all_children:
            dimensions = [
                c.preferred_width(max_available_width) for c in self._all_children
            ]
            return max_layout_dimensions(dimensions)
        else:
            return Dimension()

    def reset(self) -> "None":
        """Reset the state of this container and all the children.

        Does nothing as this container is used for dumping outpiu.
        """
        pass



# -*- coding: utf-8 -*-
"""Defines a configuration class for euporie."""
from __future__ import annotations

import argparse
import json
import logging
import os
from collections import ChainMap
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, Optional, Sequence, Union

import jsonschema  # type: ignore
from appdirs import user_config_dir  # type: ignore
from pygments.styles import get_all_styles  # type: ignore

from euporie import __app_name__, __copyright__, __strapline__, __version__

log = logging.getLogger(__name__)


class JSONEncoderPlus(json.JSONEncoder):
    """JSON encode class which encodes paths as strings."""

    def default(self, o: "Any") -> "Union[bool, int, float, str, None]":
        """Encode an object to JSON.

        Args:
            o: The object to encode

        Returns:
            The encoded object

        """
        if isinstance(o, Path):
            return str(o)
        return json.JSONEncoder.default(self, o)


_json_encoder = JSONEncoderPlus()


class BooleanOptionalAction(argparse.Action):
    """Action for boolean flags.

    Included because `argparse.BooleanOptionalAction` is not present in `python<=3.9`.
    """

    def __init__(self, option_strings: "list[str]", *args: "Any", **kwargs: "Any"):
        """Initate the Action, as per `argparse.BooleanOptionalAction`."""
        _option_strings = list(option_strings)
        for option_string in option_strings:
            if option_string.startswith("--"):
                _option_strings.append(f"--no-{option_string[2:]}")
        kwargs["nargs"] = 0
        super().__init__(_option_strings, *args, **kwargs)

    def __call__(
        self,
        parser: "argparse.ArgumentParser",
        namespace: "argparse.Namespace",
        values: "Union[str, Sequence[Any], None]",
        option_string: "Optional[str]" = None,
    ) -> "None":
        """Set the value to True or False depending on the flag provided."""
        if option_string in self.option_strings:
            assert isinstance(option_string, str)
            setattr(namespace, self.dest, not option_string.startswith("--no-"))

    def format_usage(self) -> "str":
        """Formats the action string.

        Returns:
            The formatted string.

        """
        return " | ".join(self.option_strings)


CONFIG_PARAMS: "dict[str, dict]" = {
    "version": {
        "flags_": ["--verion", "-V"],
        "action": "version",
        "version": f"%(prog)s {__version__}",
    },
    "dump": {
        "flags_": ["--dump"],
        "action": BooleanOptionalAction,
        "type": bool,
        "help": "Output formatted file to display or file",
        "schema_": {
            "type": "boolean",
            "default": False,
        },
        "description_": """
            When set, the formatted output will be written to the the output file path
            given by `dump_file` (standard output by default).
        """,
    },
    "dump_file": {
        "flags_": ["--dump-file"],
        "nargs": "?",
        "const": "-",
        "type": Path,
        "help": "Output path when dumping file",
        "schema_": {
            "type": "string",
            "default": None,
        },
        "description_": """
            When set to a file path, the formatted output will be written to the
            given path. If no value is given (or the default "-" is passed) output
            will be printed to standard output.
        """,
    },
    "page": {
        "flags_": ["--page"],
        "action": BooleanOptionalAction,
        "help": "Pass output to pager",
        "type": bool,
        "schema_": {
            "type": "boolean",
            "default": False,
        },
        "description_": """
            Whether to pipe output to the system pager when using `--dump`.
        """,
    },
    # "run": {
    # "flags_": ["--run"],
    # "action": BooleanOptionalAction,
    # "help": "Run the notebook when loaded",
    # "schema_": {
    # "type": "boolean",
    # "default": False,
    # },
    # "description_": """
    # If set, notebooks will be run automatically when opened, or if dumping
    # output, notebooks will be run before being output.
    # """,
    # },
    "key_map": {
        "flags_": ["--key-map"],
        "type": str,
        "choices": ["emacs", "vi"],
        "help": "Key-binding mode for text editing",
        "schema_": {
            "type": "string",
            "default": "emacs",
        },
        "description_": """
            Key binding mode to use when editing cells.
        """,
    },
    "run_after_external_edit": {
        "flags_": ["--run-after-external-edit"],
        "type": bool,
        "help": "Run cells after editing externally",
        "schema_": {
            "type": "boolean",
            "default": False,
        },
        "description_": """
            Whether to execute a cell immediately after editing in `$EDITOR`.
        """,
    },
    "max_notebook_width": {
        "flags_": ["--max-notebook-width"],
        "type": int,
        "help": "Maximum width of notebooks",
        "schema_": {
            "type": "integer",
            "minimum": 1,
            "default": 120,
        },
        "description_": """
            The maximum width at which to display a notebook.
        """,
    },
    "background_pattern": {
        "flags_": ["--background-pattern"],
        "type": int,
        "choices": range(5),
        "help": "The background pattern to use",
        "schema_": {
            "type": "integer",
            "minimum": 0,
            "maximum": 4,
            "default": 1,
        },
        "description_": """
            The background pattern to use when the notebook is narrower than the
            availble width. Zero mean no pattern is used.
        """,
    },
    "background_character": {
        "flags_": ["--background-character"],
        "type": str,
        "help": "Character for background pattern",
        "schema_": {
            "type": "string",
            "maxLength": 1,
            "default": "·",
        },
        "description_": """
            The character to use when drawing the background pattern.
        """,
    },
    "line_numbers": {
        "flags_": ["--line-numbers"],
        "action": BooleanOptionalAction,
        "type": bool,
        "help": "Show or hide line numbers",
        "schema_": {
            "type": "boolean",
            "default": True,
        },
        "description_": """
            Whether line numbers are shown by default.
        """,
    },
    "syntax_theme": {
        "flags_": ["--syntax-theme"],
        "type": str,
        # Do not want to print all theme names in --help screen as it looks messy
        # "choices": list(get_all_styles()),
        "help": "Syntax higlighting theme",
        "schema_": {
            "type": "string",
            "pattern": f"({'|'.join(get_all_styles())})",
            "default": "default",
        },
        "description_": """
            The name of the pygments style for syntax highlighting.
        """,
    },
    "files": {
        "flags_": ["files"],
        "nargs": "*",
        "default": [],
        "type": Path,
        "help": "List of file names to open",
        "schema_": {
            "type": "array",
            "items": {
                "file": {
                    "description": "File path",
                    "type": "string",
                }
            },
        },
        "description_": """
        """,
    },
}

CONFIG_SCHEMA: "dict" = {
    "title": "Euporie Configuration",
    "description": "A configuration for euporie",
    "type": "object",
    "properties": {
        name: {
            "description": param.get("help"),
            **(
                {"pattern": f"({'|'.join(choices)})"}
                if (choices := param.get("choices")) and param.get("type") == str
                else {}
            ),
            **({"default": default} if (default := param.get("default")) else {}),
            **param["schema_"],
        }
        for name, param in CONFIG_PARAMS.items()
        if param.get("schema_")
    },
}


class Config:
    """A configuration object with configuration values available as attributes.

    Default configuration variables are loaded from the defaults defined in the
    schema, then overwritten with values defined in a configuration file.
    """

    conf_file_name = "config.json"
    defaults = {
        name: param.get("schema_", {}).get("default")
        for name, param in CONFIG_PARAMS.items()
    }

    def __init__(self):
        """Ininitate the Configuration object."""
        self.user = {}
        self.env = {}
        self.args = {}

        user_conf_dir = Path(user_config_dir(__app_name__, appauthor=False))
        user_conf_dir.mkdir(exist_ok=True, parents=True)
        self.config_file_path = user_conf_dir / self.conf_file_name
        self.valid_user = True

        self.load_user()
        self.load_env()
        self.load_args()

        self.chain = ChainMap(
            self.args,
            self.env,
            self.user,
            self.defaults,
        )

    def load_args(self) -> "None":
        """Attempts to load configuration settings from commandline flags."""
        parser = argparse.ArgumentParser(
            description=__strapline__,
            epilog=__copyright__,
            allow_abbrev=True,
            formatter_class=argparse.MetavarTypeHelpFormatter,
        )
        for name, data in CONFIG_PARAMS.items():
            parser.add_argument(
                *data.get("flags_") or [name],
                # Do not set defaults for command line arguments, as default values
                # would override values set in the configuration file
                **{
                    key: value
                    for key, value in data.items()
                    if not key.endswith("_") and key != "default"
                },
            )
        for name, value in vars(parser.parse_args()).items():
            if value is not None:
                # Convert to json and back to attain json types
                json_data = json.loads(_json_encoder.encode({name: value}))
                try:
                    jsonschema.validate(instance=json_data, schema=CONFIG_SCHEMA)
                except jsonschema.ValidationError as error:
                    log.warning(f"Error in command line parameter `{name}`: {error}")
                else:
                    self.args[name] = value

    def load_env(self) -> "None":
        """Attempt to load configuration settings from environment variables."""
        for name, param in CONFIG_PARAMS.items():
            env = f"{__app_name__.upper()}_{name.upper()}"
            if env in os.environ:
                type_ = param.get("type", str)
                try:
                    value = type_(os.environ[env])
                except (ValueError, TypeError):
                    log.warning(
                        f"Environment variable `{env}` not understood"
                        f" - `{type_.__name__}` expected"
                    )
                else:
                    json_data = json.loads(_json_encoder.encode({name: value}))
                    try:
                        jsonschema.validate(instance=json_data, schema=CONFIG_SCHEMA)
                    except jsonschema.ValidationError as error:
                        log.error(f"Error in environment variable: `{env}`\n{error}")
                    else:
                        self.env[name] = value

    def load_user(self) -> "None":
        """Attempt to load JSON configuration file."""
        assert isinstance(self.config_file_path, Path)
        if self.valid_user and self.config_file_path.exists():
            with open(self.config_file_path, "r") as f:
                try:
                    json_data = json.load(f)
                except json.decoder.JSONDecodeError:
                    log.error(
                        "Could not parse the configuration file: "
                        f"{self.config_file_path}\n"
                        "Is it valid json?"
                    )
                    self.valid_user = False
                else:
                    try:
                        jsonschema.validate(instance=json_data, schema=CONFIG_SCHEMA)
                    except jsonschema.ValidationError as error:
                        log.warning(
                            f"Error in config file: {self.config_file_path}\n"
                            f"{error}"
                        )
                        self.valid_user = False
                    else:
                        self.user.update(json_data)
                        return
            log.warning("The configuration file was not loaded")

    def get(self, name: "str") -> "Any":
        """Access a configuration variable, falling back to the default value if unset.

        Args:
            name: The name of the attribute to access.

        Returns:
            The configuration variable value.

        """
        return self.chain.get(name)

    def __getattr__(self, name: "str") -> "Any":
        """Enables access of config elements via dotted attributes.

        Args:
            name: The name of the attribute to access.

        Returns:
            The configuration variable value.

        """
        return self.get(name)

    def __setitem__(self, name: "str", value: "Any") -> "None":
        """Set a configuration attribute.

        Args:
            name: The name of the attribute to set.
            value: The value to give the attribute.

        """
        if name in self.args:
            del self.args[name]
        self.user[name] = value
        if self.valid_user:
            with open(self.config_file_path, "w") as f:
                json.dump(self.user, f, indent=2)

    def __setattr__(self, attr: "str", value: "Union[bool, int, str]") -> "None":
        """Sets configuration attributes and writes their values to the config file."""
        if attr in self.defaults:
            self.__setitem__(attr, value)
        else:
            return super().__setattr__(attr, value)

    def __delitem__(self, name: "str") -> "None":
        """Unset a user's configuration variable.

        This removes a configuration setting from the user's configuration, so the
        default value will be used.

        Args:
            name: The name of the attribute to unset.

        Raises:
            KeyError: When the configuration does not exist in the configuration
                schema.

        """
        try:
            del self.user[name]
        except KeyError:
            raise KeyError(f"Variable not found in the user config: {name!r}")
        else:
            if self.valid_user:
                with open(self.config_file_path, "w") as f:
                    json.dump(self.user, f, indent=2)

    def __iter__(self) -> "Iterator[str]":
        """Iterate over all configuration variable names.

        Returns:
            An iterable of the combined dictionary.

        """
        return iter(self.chain)

    def __len__(self) -> "int":
        """Return the length of the combined user and default settings.

        Returns:
            The length of the combined dictionary.

        """
        return len(self.chain)

    def __str__(self) -> "str":
        """Represent the configuration as a string.

        Returns:
            A string representing the configuration.

        """
        return f"Config({self.chain!r})"

    __repr__ = __str__

    def toggle(self, name: "str") -> "None":
        """Switches attributes between permitted configuration states.

        For boolean values, they are toggled between True and False. Integer values are
        incremented and reset within the permitted range.

        Args:
            name: The name of the attribute to toggle.

        """
        if name in self.defaults:
            current = getattr(self, name)
            schema = CONFIG_SCHEMA["properties"][name]
            if schema["type"] == "boolean":
                setattr(self, name, not current)
            elif schema["type"] == "integer":
                setattr(
                    self,
                    name,
                    schema["minimum"]
                    + (current - schema["minimum"] + 1) % (schema["maximum"] + 1),
                )


# Do not actually load the config if type checking - it causes pytype to exit
config: "Config"
if not TYPE_CHECKING:
    config = Config()



# -*- coding: utf-8 -*-
"""Defines a smarter menu item."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, Union

from prompt_toolkit.widgets.menus import MenuItem

if TYPE_CHECKING:
    from prompt_toolkit.filters import Filter
    from prompt_toolkit.keys import Keys


class SmartMenuItem(MenuItem):
    """A prompt-toolkit compatible menu item with more advanced capabilities.

    It can use a function to generate the text to display, display a checkmark if a
    condition is true, and disable the handler if a condition is met.
    """

    def __init__(
        self,
        text: "Union[str, Callable]" = "",
        handler: "Optional[Callable[[], None]]" = None,
        children: "Optional[list[MenuItem]]" = None,
        shortcut: "Optional[Sequence[Union[Keys, str]]]" = None,
        disabler: "Union[Filter]" = None,
        toggler: "Optional[Filter]" = None,
    ) -> None:
        """Initiate a smart menu item.

        Args:
            text: A string, or a callable which returns the text to display.
            handler: As per `prompt_toolkit.widgets.menus.MenuItem`.
            children: As per `prompt_toolkit.widgets.menus.MenuItem`.
            shortcut: As per `prompt_toolkit.widgets.menus.MenuItem`.
            disabler: The handler will be disabled when this callable returns True.
            toggler: A checkmark will be displayed next to the menu text when this
                callable returns true.

        """
        self.text_generator = text
        self.disabler = disabler
        self.toggler = toggler
        self._handler = handler
        super().__init__(self.text, handler, children, shortcut, False)

    # Type checking disabled for the following property methods due to open mypy bug:
    # https://github.com/python/mypy/issues/4125

    @property  # type: ignore
    def text(self) -> "str":  # type: ignore
        """Generate the text for this menu item."""
        if callable(self.text_generator):
            text = self.text_generator()
        else:
            text = self.text_generator
        if self.toggler is not None:
            text += " ✓" if self.toggler() else ""

        # Check if this menu item should be disabled, and if so, remove the handler
        self.handler = None if self.disabled else self._handler

        return text

    @text.setter
    def text(self, value: "Any") -> "None":
        """Prevent the inherited `__init__` method setting this property value."""
        pass

    @property  # type: ignore
    def disabled(self) -> "bool":  # type: ignore
        """Determine if the menu item is disabled."""
        return self.disabler is not None and self.disabler()

    @disabled.setter
    def disabled(self, value: "Any") -> None:
        """Prevent the inherited `__init__` method setting this property value."""
        pass



# -*- coding: utf-8 -*-
"""Contains the main Application class which runs euporie."""
from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING, Callable

from prompt_toolkit.clipboard import Clipboard, InMemoryClipboard
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import Condition
from prompt_toolkit.key_binding.bindings.focus import focus_next, focus_previous
from prompt_toolkit.layout import (
    ConditionalContainer,
    DynamicContainer,
    Float,
    FloatContainer,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import LayoutDimension as D
from prompt_toolkit.layout.menus import CompletionsMenu
from prompt_toolkit.output.defaults import create_output
from prompt_toolkit.widgets import MenuContainer, MenuItem
from pygments.styles import get_all_styles  # type: ignore
from pyperclip import determine_clipboard  # type: ignore

from euporie import __logo__
from euporie.app.dialog import DialogMixin
from euporie.config import config
from euporie.keys import KeyBindingsInfo
from euporie.menu import SmartMenuItem
from euporie.notebook import File, Notebook
from euporie.scroll import ScrollingContainer

if TYPE_CHECKING:
    from prompt_toolkit.key_binding import KeyBindings
    from prompt_toolkit.key_binding.key_processor import KeyPressEvent
    from prompt_toolkit.layout.containers import AnyContainer

log = logging.getLogger(__name__)


class InterfaceMixin(DialogMixin):
    """Provides user interface."""

    pre_run: "list[Callable]"
    file_op: "Callable"
    files: "list[File]"
    file: "AnyContainer"
    is_file_open: "Condition"
    close_file: "Callable"
    update_style: "Callable"
    cleanup_closed_file: "Callable"
    exit: "Callable"
    # output: "Optional[Output]"

    def setup(self) -> "None":
        """Create an output and set pre-run commands."""
        # Create an output
        self.output = create_output()
        # Ensure a file is focused if one has been opened
        self.pre_run.append(lambda: self.file_op("focus"))
        # Set clipboard to use pyperclip if possible
        self.clipboard: "Clipboard"
        if determine_clipboard()[0]:
            self.clipboard = PyperclipClipboard()
        else:
            self.clipboard = InMemoryClipboard()

    def file_container(self) -> "AnyContainer":
        """Returns a container with all opened files.

        A 1 column window is added to the left as padding.

        Returns:
            A vertical split containing the opened file containers.

        """
        return VSplit([Window(width=1), *self.files])

    def layout_container(self) -> "AnyContainer":
        """Builds the main application layout."""
        kernel_status_repr = {
            "starting": "◍",
            "idle": "○",
            "busy": "●",
            "error": "☹",
        }

        def get_statusbar_text() -> "list[tuple[str, str]]":
            """Generates the formatted text for the statusbar."""
            file = self.file
            assert isinstance(file, Notebook)
            kernel_status = kernel_status_repr.get(file.kernel_status, "◌")
            assert isinstance(file.page, ScrollingContainer)
            selected_cell = file.page.selected_index + 1
            dirt = "*" if file.dirty else ""
            return [
                ("class:menu-bar.item", f" Cell {selected_cell} "),
                ("", " "),
                ("class:menu-bar.item bold", f" {dirt}{file.path.name} "),
                ("", " "),
                ("class:menu-bar.item", f" {file.kernel_name} {kernel_status} "),
            ]

        status_bar_text = FormattedTextControl(
            get_statusbar_text, focusable=True, show_cursor=False
        )
        self.status_bar = ConditionalContainer(
            Window(
                content=status_bar_text,
                height=D.exact(1),
                style="class:menu-bar",
                dont_extend_width=True,
                align=WindowAlign.RIGHT,
            ),
            filter=self.is_file_open,
        )

        self.body_container = MenuContainer(
            body=DynamicContainer(self.file_container),
            menu_items=[
                MenuItem(
                    " File ",
                    children=[
                        MenuItem(
                            "New", handler=lambda: self.ask_open_file(validate=False)
                        ),
                        MenuItem("Open", handler=self.ask_open_file),
                        MenuItem("-", disabled=True),
                        SmartMenuItem(
                            "Save",
                            handler=lambda: self.file_op("save"),
                            disabler=~self.is_file_open,
                        ),
                        SmartMenuItem(
                            "Close",
                            handler=self.close_file,
                            disabler=~self.is_file_open,
                        ),
                        MenuItem("-", disabled=True),
                        MenuItem("Exit", handler=self.try_to_exit),
                    ],
                ),
                MenuItem(
                    " Edit ",
                    children=[
                        SmartMenuItem(
                            "Cut Cell",
                            handler=lambda: self.file_op("cut"),
                            disabler=~self.is_file_open,
                        ),
                        SmartMenuItem(
                            "Copy Cell",
                            handler=lambda: self.file_op("copy"),
                            disabler=~self.is_file_open,
                        ),
                        SmartMenuItem(
                            "Paste Cell",
                            handler=lambda: self.file_op("paste"),
                            disabler=~self.is_file_open,
                        ),
                    ],
                ),
                MenuItem(
                    " Kernel ",
                    children=[
                        MenuItem(
                            "Restart Kernel",
                            handler=lambda: self.file_op("restart_kernel"),
                        ),
                        MenuItem(
                            "Change Kernel...",
                            handler=lambda: self.file_op("change_kernel"),
                        ),
                    ],
                ),
                MenuItem(
                    " Settings ",
                    children=[
                        MenuItem(
                            "Syntax Theme",
                            children=[
                                SmartMenuItem(
                                    style,
                                    handler=partial(self.update_style, style),
                                    toggler=Condition(
                                        partial(
                                            lambda x: config.syntax_theme == x, style
                                        )
                                    ),
                                )
                                for style in sorted(get_all_styles())
                            ],
                        ),
                        MenuItem(
                            "Editing Key Bindings",
                            children=[
                                SmartMenuItem(
                                    "Emacs",
                                    handler=lambda: self.set_edit_mode("emacs"),
                                    toggler=Condition(
                                        lambda: config.key_map == "emacs"
                                    ),
                                ),
                                SmartMenuItem(
                                    "Vi",
                                    handler=lambda: self.set_edit_mode("vi"),
                                    toggler=Condition(lambda: config.key_map == "vi"),
                                ),
                            ],
                        ),
                        MenuItem(
                            "Switch Background Pattern",
                            handler=lambda: config.toggle("background_pattern"),
                        ),
                        SmartMenuItem(
                            "Run cell after external edit",
                            toggler=Condition(
                                lambda: bool(config.run_after_external_edit)
                            ),
                            handler=lambda: config.toggle("run_after_external_edit"),
                        ),
                    ],
                ),
                MenuItem(
                    " Help ",
                    children=[
                        MenuItem("Keyboard Shortcuts", handler=self.help_keys),
                        MenuItem("Logs", handler=self.help_logs),
                        MenuItem("About", handler=self.help_about),
                    ],
                ),
            ],
        )

        self.logo = FormattedTextControl(
            [("", f" {__logo__} ")],
            focusable=True,
            show_cursor=False,
            style="class:menu-bar.item,logo",
        )

        # Add logo and status bar into menubar
        assert isinstance(self.body_container.container.content, HSplit)
        self.body_container.container.content.children = [
            VSplit(
                [
                    Window(self.logo, height=D.exact(1), dont_extend_width=True),
                    self.body_container.window,
                    self.status_bar,
                ]
            ),
            self.body_container.container.content.children[1],
        ]

        self.root_container = FloatContainer(
            content=self.body_container,
            floats=[
                Float(
                    xcursor=True,
                    ycursor=True,
                    content=CompletionsMenu(max_height=16, scroll_offset=1),
                )
            ],
        )

        return self.root_container

    def set_edit_mode(self, mode: "str") -> "None":
        """Sets the keybindings for editing mode.

        Args:
            mode: 'vi' or 'emacs'

        """
        config.key_map = mode
        self.editing_mode = self.get_edit_mode()

    def get_edit_mode(self) -> "EditingMode":
        """Returns the editing mode enum defined in the configuration."""
        return {"emacs": EditingMode.EMACS, "vi": EditingMode.VI}.get(
            str(config.key_map), EditingMode.EMACS
        )

    def load_key_bindings(self) -> "KeyBindings":
        """Define application-wide keybindings."""
        kb = KeyBindingsInfo()

        @Condition
        def in_edit_mode() -> "bool":
            assert isinstance(self.file, File)
            file = self.file
            if file:
                cell = getattr(self.file, "cell", False)
                if cell:
                    if cell.is_editing():
                        return True
            return False

        @kb.add("c-n", group="Application", desc="Create a new notebook file")
        def new(event: "KeyPressEvent") -> None:
            self.ask_open_file(validate=False)

        @kb.add("c-o", group="Application", desc="Open file")
        def open(event: "KeyPressEvent") -> None:
            self.ask_open_file()

        @kb.add("c-w", group="Application", desc="Close the current file")
        def close(event: "KeyPressEvent") -> None:
            self.close_file(self.file)

        @kb.add("c-q", group="Application", desc="Quit euporie")
        def exit(event: "KeyPressEvent") -> None:
            self.try_to_exit()

        kb.add(
            "tab", filter=~in_edit_mode, group="Navigation", desc="Focus next element"
        )(focus_next)
        kb.add(
            "s-tab",
            filter=~in_edit_mode,
            group="Navigation",
            desc="Focus previous element",
        )(focus_previous)

        return kb

    def try_to_exit(self) -> "None":
        """Check for unsaved files before closing.

        Creates a chain of close file commands, where the callback for each triggers
        the closure of the next. The closing process can be cancelled anywhere along
        the chain.
        """
        if self.files:

            def final_cb() -> "None":
                """Really exit after the last file in the chain is closed."""
                self.cleanup_closed_file(self.files[0])
                self.exit()

            def create_cb(
                close_file: "File", cleanup_file: "File", cb: "Callable"
            ) -> "Callable":
                """Generate a file close chaining callbacks.

                Cleans up after the previously closed file, and requests to close the
                next file in the chain.

                Args:
                    close_file: The file to close
                    cleanup_file: The previously closed file to cleanup
                    cb: The callback to call when work is complete

                Returns:
                    A callback function which cleans up `cleanup_file` and closes
                        `close_file`.

                """

                def inner() -> None:
                    self.cleanup_closed_file(cleanup_file)
                    close_file.close(cb=cb)

                return inner

            cb = final_cb
            for close_file, cleanup_file in zip(self.files, self.files[1:]):
                cb = create_cb(close_file, cleanup_file, cb)
            self.files[-1].close(cb)
        else:
            self.exit()



# -*- coding: utf-8 -*-
"""Initiate logging for euporie."""
from __future__ import annotations

import logging
import logging.config
from tempfile import TemporaryFile
from typing import IO, cast

from prompt_toolkit.patch_stdout import StdoutProxy
from rich.console import Console

log_stdout = cast("IO[str]", StdoutProxy(raw=True))
log_memory = TemporaryFile(mode="w+")

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "plain_format": {
            "format": "{asctime} {levelname:>7} [{name}.{funcName}:{lineno}] {message}",
            "style": "{",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
        "rich_format": {
            "format": "{message}",
            "style": "{",
            "datefmt": "%Y-%m-%d %H:%M:%S",
        },
    },
    "handlers": {
        "file": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "plain_format",
        },
        "memory": {
            "level": "INFO",
            "class": "rich.logging.RichHandler",
            "console": Console(file=log_memory, force_terminal=True, width=120),
            "markup": True,
            "rich_tracebacks": True,
            "formatter": "rich_format",
        },
        "screen": {
            "level": "ERROR",
            "class": "rich.logging.RichHandler",
            "console": Console(file=log_stdout, force_terminal=True),
            "markup": True,
            "rich_tracebacks": True,
            "formatter": "rich_format",
        },
    },
    "loggers": {
        "euporie": {
            "handlers": ["memory", "screen"],
            "level": "DEBUG",
            "propagate": False,
        },
    },
    "root": {"handlers": ["memory"]},
}

logging.config.dictConfig(LOGGING_CONFIG)



# -*- coding: utf-8 -*-
"""Contains the main Application class which runs euporie."""
from __future__ import annotations

import logging
import sys
from asyncio import AbstractEventLoop
from pathlib import Path
from typing import Any, Callable, Optional

from prompt_toolkit.formatted_text import (
    AnyFormattedText,
    fragment_list_to_text,
    to_formatted_text,
)
from prompt_toolkit.layout import Float, HSplit, Layout, Window
from prompt_toolkit.layout.containers import AnyContainer, FloatContainer, to_container
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.layout.dimension import LayoutDimension as D
from prompt_toolkit.widgets import Button, Dialog, Label, TextArea

from euporie import __app_name__, __copyright__, __logo__, __strapline__, __version__
from euporie.keys import KeyBindingsInfo
from euporie.log import log_memory
from euporie.text import ANSI, FormattedTextArea

log = logging.getLogger(__name__)


class DialogMixin:
    """Provides dialogs for the main application."""

    root_container: "FloatContainer"
    layout: "Layout"
    open_file: "Callable"

    def dialog(
        self,
        title: "AnyFormattedText",
        body: "AnyContainer",
        buttons: "dict[str, Optional[Callable]]",
        to_focus: "Optional[AnyContainer]" = None,
    ) -> None:
        """Display a modal dialog above the application.

        Returns focus to the previously selected control when closed.

        Args:
            title: The title of the dialog. Can be formatted text.
            body: The container to use as the main body of the dialog.
            buttons: A dictionary mapping text to display as dialog buttons to
                callbacks to run when the button is clicked. If the callback is
                `None`, the dialog will be closed without running a callback.
            to_focus: The control to focus when the dialog is displayed.

        """

        def _make_handler(cb: "Optional[Callable]") -> "Callable":
            def inner() -> "None":
                self.root_container.floats.remove(dialog)
                self.layout.focus(focused)
                if cb:
                    cb()

            return inner

        focused = self.layout.current_control

        button_widgets = [
            Button(text, _make_handler(cb), left_symbol="[", right_symbol="]")
            for text, cb in buttons.items()
        ]

        dialog = Float(
            Dialog(
                title=title,
                body=body,
                buttons=button_widgets,
                modal=True,
                with_background=True,
            )
        )
        self.root_container.floats.append(
            dialog,
        )
        if to_focus is None:
            to_focus = button_widgets[0]
        self.layout.focus(to_focus)

    def ask_open_file(
        self,
        default: "str" = "",
        validate: "bool" = True,
        error: "Optional[str]" = None,
    ) -> None:
        """Display a dialog asking for file name input.

        Args:
            default: The default filename to display in the text entry box
            validate: Whether to disallow files which do not exist
            error: An optional error message to display below the file name

        """
        filepath = TextArea(text=default, multiline=False)

        def _open_cb() -> None:
            path = filepath.text
            if not validate or Path(path).expanduser().exists():
                self.open_file(filepath.text)
            else:
                self.ask_open_file(
                    default=filepath.text, validate=validate, error="File not found"
                )

        body_contents: "list[AnyContainer]" = [
            Label("Enter file name:"),
            filepath,
        ]
        if error:
            body_contents.append(to_container(Label(error, style="red")))
        self.dialog(
            title="Select file",
            body=HSplit(body_contents),
            buttons={
                "OK": _open_cb,
                "Cancel": None,
            },
            to_focus=filepath,
        )

    def help_keys(self) -> None:
        """Displays details of registered key-bindings in a dialog."""
        key_details = KeyBindingsInfo.to_formatted_text()
        max_line_width = max(
            [len(line) for line in fragment_list_to_text(key_details).split("\n")]
        )
        body = FormattedTextArea(
            formatted_text=key_details,
            multiline=True,
            focusable=True,
            wrap_lines=False,
            width=D(preferred=max_line_width + 2),
            scrollbar=True,
        )

        self.dialog(
            title="Keyboard Shortcuts",
            body=body,
            buttons={"OK": None},
        )

    def help_logs(self) -> None:
        """Displays a dialog with logs."""
        log_memory.seek(0)
        log_data = to_formatted_text(ANSI(log_memory.read()))

        body = FormattedTextArea(
            formatted_text=log_data,
            multiline=True,
            focusable=True,
            wrap_lines=False,
            width=D(preferred=120),
            scrollbar=True,
        )
        self.dialog(
            title="Logs",
            body=body,
            buttons={"OK": None},
        )

    def help_about(self) -> None:
        """Displays an about dialog."""
        self.dialog(
            title="About",
            body=Window(
                FormattedTextControl(
                    [
                        ("class:logo", __logo__),
                        ("", " "),
                        ("bold", __app_name__),
                        ("", f"Version {__version__}\n\n".rjust(27, " ")),
                        ("", __strapline__),
                        ("", "\n"),
                        ("class:hr", "─" * 34 + "\n\n"),
                        ("", __copyright__),
                    ]
                ),
                dont_extend_height=True,
            ),
            buttons={"OK": None},
        )

    def _handle_exception(
        self, loop: "AbstractEventLoop", context: "dict[str, Any]"
    ) -> None:
        from prompt_toolkit.formatted_text import to_formatted_text
        from rich.traceback import Traceback

        from euporie.render import RichRenderer

        # Log observed exceptions to the log
        log.error("An unhandled exception occured", exc_info=sys.exc_info())

        # Also display a dialog to the user
        formatted_tb = to_formatted_text(
            ANSI(RichRenderer().render(Traceback(), width=80, height=999))
        )
        tb_control = FormattedTextArea(
            formatted_text=formatted_tb,
            multiline=True,
            focusable=True,
            wrap_lines=False,
            width=D(preferred=81),
            height=D(max=15),
            scrollbar=True,
        )
        self.dialog(
            title="Error",
            body=HSplit(
                [
                    Label("An error occured:\n", style="bold"),
                    tb_control,
                ]
            ),
            buttons={"OK": None},
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['euporie', 'euporie.app', 'euporie.libs', 'euporie.libs.commonmark_extensions']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.2.0,<9.0.0',
 'appdirs>=1.4.4,<2.0.0',
 'jsonschema>=3.2.0,<4.0.0',
 'jupyter-client==6.1.12',
 'nbformat>=5.1.3,<6.0.0',
 'prompt-toolkit>=3.0.18,<4.0.0',
 'pyperclip>=1.8.2,<2.0.0',
 'rich>=10.1.0,<11.0.0']

extras_require = \
{'html-mtable': ['mtable>=0.1.10,<0.2.0', 'html5lib>=1.1,<2.0'],
 'images-img2unicode': ['img2unicode>=0.1a7,<0.2'],
 'images-timg': ['timg>=1.1.5,<2.0.0']}

entry_points = \
{'console_scripts': ['euporie = euporie.app:App.launch']}

setup_kwargs = {
    'name': 'euporie',
    'version': '0.1.15',
    'description': 'Euporie is a text-based user interface for running and editing Jupyter notebooks',
    'long_description': '<h1 align="center">euporie</h1>\n\n<img src="https://user-images.githubusercontent.com/12154190/117550683-79526700-b039-11eb-8a83-1828c6ee8125.png" alt="screenshot 1" width="45%" align="right" />\n\n# About\n\nEuporie is a text-based user interface for running and editing Jupyter notebooks.\n\n# Install\n\nEuporie is on pypi, so can be installed using `pip` or [`pipx`](https://pipxproject.github.io/pipx/):\n\n```bash\n# install inside a virtualenv\npip install euporie\n\n# install globally\npipx install euporie\n```\n\nYou may wish to install some optional python dependencies to render images and HTML tables (but [see below](#image-support)):\n\n```bash\npip install euporie[html-mtable,images-timg]\n```\n\n# Screenshots\n\n<p align="center">\n<img src="https://user-images.githubusercontent.com/12154190/117550685-7a839400-b039-11eb-98ac-8adb9ea2cfc3.png" alt="screenshot 2" width="19%" />\n<img src="https://user-images.githubusercontent.com/12154190/117550686-7a839400-b039-11eb-8c6f-65c3cedf2f25.png" alt="screenshot 3" width="19%" />\n<img src="https://user-images.githubusercontent.com/12154190/117550687-7b1c2a80-b039-11eb-867b-d5e9d8671495.png" alt="screenshot 4" width="19%" />\n<img src="https://user-images.githubusercontent.com/12154190/117550688-7bb4c100-b039-11eb-9419-a10c8c0f9b21.png" alt="screenshot 5" width="19%" />\n<img src="https://user-images.githubusercontent.com/12154190/117550689-7bb4c100-b039-11eb-9d90-44df4c0e0f03.png" alt="screenshot 6" width="19%" />\n</p>\n\n# Features\n\n- Execute notebooks in the terminal\n- Autocompletion in code cells\n- Rich output support, including:\n  - Markdown\n  - Tables\n  - Images\n- Open multiple notebooks side-by-side\n- Good performance with large notebook files\n\n## Image Support\n\nEuporie will attempt to render images in the best possible way it can. The following methods are supported:\n\n- **Sixel**\n\n  If supported by your terminal, euporie can show graphical images in cell outputs\n  This requires one of the following:\n\n  [`imagemagik`](https://www.imagemagick.org)\n  [`timg`](https://github.com/adzierzanowski/timg)\n  [`teimpy`](https://github.com/ar90n/teimpy)\n\n- **Kitty\'s terminal image protocol**\n\n  If your terminal supports [kitty\'s terminal graphics protocol](https://sw.kovidgoyal.net/kitty/graphics-protocol.html), euporie will use it to render images\n\n- **Ansi art**\n\n  This requires one of the following:\n\n  [`timg`](https://github.com/hzeller/timg)\n  [`catimg`](https://github.com/posva/catimg)\n  [`icat`](https://github.com/atextor/icat)\n  [`timg`](https://github.com/adzierzanowski/timg)\n  [`tiv`](https://github.com/radare/tiv)\n  [`viu`](https://github.com/atanunq/viu)\n  [`img2unicode`](https://github.com/matrach/img2unicode)\n  [`jp2a`](https://csl.name/jp2a/)\n  [`img2txt`](http://caca.zoy.org/wiki/libcaca)\n\nThe kitty & sixel image rendering methods will fall back to ansi art images when rendering images in partially obscured cells, to prevent clipped images destroying the user interface.\n\nFor SVG support, [`cairosvg`](https://cairosvg.org/) or [`imagemagik`](https://www.imagemagick.org) are required.\n\n## HTML Support\n\nEuporie will attempt to render HTML outputs. This requires one of the following:\n\n[`w3m`](http://w3m.sourceforge.net/)\n[`elinks`](http://elinks.or.cz/)\n[`lynx`](https://lynx.browser.org/)\n[`links`](http://links.twibright.com/)\n[`mtable`](https://github.com/liuyug/mtable)\n\n_Note: only HTML tables will be displayed if `mtable` is used_\n\nIf none of these commands are found in your `$PATH`, the plain text representation will be used.\n\n# Usage\n\n```\nusage: euporie [-h] [-V] [--dump | --no-dump] [--execute | --no-execute]\n               [--editing-mode {emacs,vi}]\n               [--execute-after-external-edit | --no-execute-after-external-edit]\n               [--max-notebook-width int] [--background {0,1,2,3,4}]\n               [--background-character str]\n               [--show-line-numbers | --no-show-line-numbers]\n               [--pygments-style]\n               [Path ...]\n\npositional arguments:\n  Path                  List of file names to open\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -V, --version         show program\'s version number and exit\n\nrun options:\n  --dump, --no-dump     Whether to print the rendered files to the terminal\n                        then exit\n  --execute, --no-execute\n                        Whether to execute the notebook when it is loaded\n\ninteractive mode options:\n  --editing-mode {emacs,vi}\n                        The key-binding style to use for text editing\n  --execute-after-external-edit, --no-execute-after-external-edit\n                        Whether to execute a cell immediately after editing in\n                        $EDITOR\n\ndisplay options:\n  --max-notebook-width int\n                        The maximum width of the notebook\n  --background {0,1,2,3,4}\n                        The background pattern to use\n  --background-character str\n                        The character to use to draw the background\n  --show-line-numbers, --no-show-line-numbers\n                        Whether line numbers are shown by default\n  --pygments-style\n                        The name of the pygments style for syntax highlighting\n```\n\n# Key Bindings\n\n|              Key Binding | Command                       |\n| -----------------------: | :---------------------------- |\n|          **Application** |                               |\n|                 `ctrl-n` | Create a new notebook file    |\n|                 `ctrl-o` | Open file                     |\n|                 `ctrl-w` | Close the current file        |\n|                 `ctrl-q` | Quit euporie                  |\n|                 `ctrl-s` | Save current file             |\n|           **Navigation** |                               |\n|                    `tab` | Focus next element            |\n|              `shift-tab` | Focus previous element        |\n|                      `[` | Scroll up                     |\n|                      `]` | Scroll down                   |\n|       `ctrl-up` / `home` | Go to first cell              |\n|                 `pageup` | Go up 5 cells                 |\n|               `up` / `k` | Go up one cell                |\n|             `down` / `j` | Go down one cell              |\n|               `pagedown` | Go down 5 cells               |\n|      `ctrl-down` / `end` | Go to last cell               |\n|             **Notebook** |                               |\n|                      `a` | Add new cell above            |\n|                      `b` | Add new cell below            |\n|                    `d d` | Delete current cell           |\n|                      `x` | Cut current cell              |\n|                      `c` | Copy current cell             |\n|                      `v` | Paste copied cell             |\n|                      `m` | Change cell to markdown       |\n|                      `y` | Change cell to code           |\n|                      `r` | Change cell to raw            |\n|                      `l` | Toggle line numbers           |\n|                      `e` | Edit cell in \\$EDITOR         |\n|                  `enter` | Enter cell edit mode          |\n|                 `escape` | Exit cell edit mode\\*         |\n|          `escape escape` | Exit cell edit mode quickly   |\n|  `ctrl-enter` / `ctrl-e` | Run cell\\*\\*                  |\n| `shift-enter` / `ctrl-r` | Run then select next cell\\*\\* |\n|            **Edit Mode** |                               |\n|                 `ctrl-f` | Find                          |\n|                 `ctrl-g` | Find Next                     |\n|                 `ctrl-z` | Undo                          |\n|                 `ctrl-d` | Duplicate line                |\n|                    `tab` | Indent                        |\n|              `shift-tab` | Unindent                      |\n|                 `ctrl-c` | Copy                          |\n|                 `ctrl-x` | Cut                           |\n|                 `ctrl-v` | Paste                         |\n\n> \\* There is a slight delay detecting an escape key-event. To exit edit mode quickly, double-press the escape key.\n>\n> \\*\\* These entries require your terminal to support CSI-u mode. If your terminal does not support this, it may be possible to work around this by remapping the keys in your terminal emulator - see below).\n\nWhen in edit mode, emacs style key-bindings apply.\n\n## Key Remapping\n\nBy default, VT100 terminal emulators do not distinguish between `Enter`, `Ctrl + Enter` & `Shift + Enter`. In order to work around this, it is possible to re-map these key bindings so they produce the escape code of another key. To replicate the `Ctrl + Enter` & `Shift + Enter` of Jupyter, you will need to remap the following shortcuts in your terminal:\n\n| Key Combination | Output       |\n| --------------- | ------------ |\n| `Ctrl + Enter`  | `Ctrl + F20` |\n| `Shift + Enter` | `F21`        |\n\n### xterm\n\nAdd the following to your `~/.Xresources`\n\n```\n*.vt100.translations: #override \\n\\\n    Ctrl <Key>Return: string("\\033\\[19;6~") \\n\\\n    Shift <Key>Return: string("\\033\\[20;2~") \\n\\\n```\n\n### konsole\n\nIn the menu, navigate to:\n\n`Settings` -> `Edit Current Profile` -> `Keyboard` -> `Edit`\n\nChange the existing entry for `Return+Shift` to `Return+Shift+Ctrl` (or whatever you prefer), then add the following entries:\n\n| Key combination | Output      |\n| --------------- | ----------- |\n| `Return+Ctrl`   | `\\E\\[19;6~` |\n| `Return+Shift`  | `\\E\\[20;2~` |\n\n# Roadmap\n\n- Add ability to dump formatted notebooks\n- Add command line argument parsing\n- Render outputs asynchronously in a separate thread\n- Upstream markdown tables in `rich`\n- Cell attachments\n- LaTeX\n- Widgets\n\n# Related projects\n\n- https://github.com/davidbrochart/nbterm\n\n  An alternative effort sponsored by QuantStack\n\n- https://github.com/chentau/nbtui\n\n  A cli Jupyter notebook viewer with support for kitty\'s terminal graphics protocol\n\n- https://github.com/mosiman/jupytui\n\n  A cli Jupyter notebook viewer\n\n- https://kracekumar.com/post/jut/\n\n  Another cli Jupyter notebook viewer\n',
    'author': 'Josiah Outram Halstead',
    'author_email': 'josiah@halstead.email',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/joouha/euporie',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
